# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the task panel code for the Draft OrthoArray tool."""
## @package task_orthoarray
# \ingroup drafttaskpanels
# \brief Provides the task panel code for the Draft OrthoArray tool.

## \addtogroup drafttaskpanels
# @{
import PySide.QtGui as QtGui
from PySide.QtCore import QT_TRANSLATE_NOOP
from PySide import QtWidgets
import FreeCAD as App
import FreeCADGui as Gui
import Draft_rc  # include resources, icons, ui files
import DraftVecUtils
from FreeCAD import Units as U
from draftutils import params
from draftutils.messages import _err, _log, _msg
from draftutils.translate import translate

# The module is used to prevent complaints from code checkers (flake8)
bool(Draft_rc.__name__)


def _quantity(st):
    # workaround for improper handling of plus sign
    # in Building US unit system
    # https://github.com/FreeCAD/FreeCAD/issues/11345
    return U.Quantity(st.replace("+", "--")).Value


class TaskPanelOrthoArray:
    """TaskPanel code for the OrthoArray command.

    The names of the widgets are defined in the `.ui` file.
    This `.ui` file `must` be loaded into an attribute
    called `self.form` so that it is loaded into the task panel correctly.

    In this class all widgets are automatically created
    as `self.form.<widget_name>`.

    The `.ui` file may use special FreeCAD widgets such as
    `Gui::InputField` (based on `QLineEdit`) and
    `Gui::QuantitySpinBox` (based on `QAbstractSpinBox`).
    See the Doxygen documentation of the corresponding files in `src/Gui/`,
    for example, `InputField.h` and `QuantitySpinBox.h`.

    Attributes
    ----------
    source_command: gui_base.GuiCommandBase
        This attribute holds a reference to the calling class
        of this task panel.
        This parent class, which is derived from `gui_base.GuiCommandBase`,
        is responsible for calling this task panel, for installing
        certain callbacks, and for removing them.

        It also delays the execution of the internal creation commands
        by using the `draftutils.todo.ToDo` class.

    See Also
    --------
    * https://forum.freecad.org/viewtopic.php?f=10&t=40007
    * https://forum.freecad.org/viewtopic.php?t=5374#p43038
    """

    def __init__(self):

        self.form = Gui.PySideUic.loadUi(":/ui/TaskPanel_OrthoArray.ui")
        self.form.setWindowTitle(translate("draft", "Orthogonal Array"))
        self.form.setWindowIcon(QtGui.QIcon(":/icons/Draft_Array.svg"))

        # -------------------------------------------------------------------
        # Default values for the internal function,
        # and for the task panel interface
        start_x = params.get_param("XInterval", "Mod/Draft/OrthoArrayLinearMode")
        start_y = params.get_param("YInterval", "Mod/Draft/OrthoArrayLinearMode")
        start_z = params.get_param("ZInterval", "Mod/Draft/OrthoArrayLinearMode")

        self.v_x = App.Vector(start_x, 0, 0)
        self.v_y = App.Vector(0, start_y, 0)
        self.v_z = App.Vector(0, 0, start_z)

        self.form.input_X_x.setProperty("rawValue", self.v_x.x)
        self.form.input_X_y.setProperty("rawValue", self.v_x.y)
        self.form.input_X_z.setProperty("rawValue", self.v_x.z)

        self.form.input_Y_x.setProperty("rawValue", self.v_y.x)
        self.form.input_Y_y.setProperty("rawValue", self.v_y.y)
        self.form.input_Y_z.setProperty("rawValue", self.v_y.z)

        self.form.input_Z_x.setProperty("rawValue", self.v_z.x)
        self.form.input_Z_y.setProperty("rawValue", self.v_z.y)
        self.form.input_Z_z.setProperty("rawValue", self.v_z.z)

        self.n_x = params.get_param("XNumOfElements", "Mod/Draft/OrthoArrayLinearMode")
        self.n_y = params.get_param("YNumOfElements", "Mod/Draft/OrthoArrayLinearMode")
        self.n_z = params.get_param("ZNumOfElements", "Mod/Draft/OrthoArrayLinearMode")

        self.form.spinbox_n_X.setValue(self.n_x)
        self.form.spinbox_n_Y.setValue(self.n_y)
        self.form.spinbox_n_Z.setValue(self.n_z)

        self.fuse = params.get_param("Draft_array_fuse")
        self.use_link = params.get_param("Draft_array_Link")

        self.form.checkbox_fuse.setChecked(self.fuse)
        self.form.checkbox_link.setChecked(self.use_link)
        # -------------------------------------------------------------------

        # Initialize Linear mode variables
        self.linear_mode = params.get_param("LinearModeOn", "Mod/Draft/OrthoArrayLinearMode")
        self.active_axis = params.get_param("AxisSelected", "Mod/Draft/OrthoArrayLinearMode")

        # Hide the axis radiobuttons initially (they're only visible in Linear mode)
        self.toggle_axis_radiobuttons(False)
        if self.linear_mode:
            self.form.button_linear_mode.setChecked(True)
            self.toggle_linear_mode()
        else:
            self.form.group_linearmode.hide()
        # -------------------------------------------------------------------

        # Some objects need to be selected before we can execute the function.
        self.selection = None

        # This is used to test the input of the internal function.
        # It should be changed to True before we can execute the function.
        self.valid_input = False

        self.set_widget_callbacks()

        self.tr_true = QT_TRANSLATE_NOOP("Draft", "True")
        self.tr_false = QT_TRANSLATE_NOOP("Draft", "False")

    def set_widget_callbacks(self):
        """Set up the callbacks (slots) for the widget signals."""
        # New style for Qt5
        self.form.button_reset_X.clicked.connect(lambda: self.reset_v("X"))
        self.form.button_reset_Y.clicked.connect(lambda: self.reset_v("Y"))
        self.form.button_reset_Z.clicked.connect(lambda: self.reset_v("Z"))

        # When the checkbox changes, change the internal value
        if hasattr(self.form.checkbox_fuse, "checkStateChanged"):  # Qt version >= 6.7.0
            self.form.checkbox_fuse.checkStateChanged.connect(self.set_fuse)
            self.form.checkbox_link.checkStateChanged.connect(self.set_link)
        else:  # Qt version < 6.7.0
            self.form.checkbox_fuse.stateChanged.connect(self.set_fuse)
            self.form.checkbox_link.stateChanged.connect(self.set_link)

        # Linear mode callbacks - only set up if the UI elements exist
        self.form.button_linear_mode.clicked.connect(self.toggle_linear_mode)
        self.form.radiobutton_x_axis.clicked.connect(lambda: self.set_active_axis("X"))
        self.form.radiobutton_y_axis.clicked.connect(lambda: self.set_active_axis("Y"))
        self.form.radiobutton_z_axis.clicked.connect(lambda: self.set_active_axis("Z"))

    def accept(self):
        """Execute when clicking the OK button or Enter key."""
        self.selection = Gui.Selection.getSelection()

        (self.v_x, self.v_y, self.v_z) = self.get_intervals()

        (self.n_x, self.n_y, self.n_z) = self.get_numbers()

        self.valid_input = self.validate_input(
            self.selection, self.v_x, self.v_y, self.v_z, self.n_x, self.n_y, self.n_z
        )
        if self.valid_input:
            axis_values = {
                "X": (self.v_x.x, self.n_x),
                "Y": (self.v_y.y, self.n_y),
                "Z": (self.v_z.z, self.n_z),
            }

            values = axis_values.get(self.active_axis)
            if values is not None:
                interval, num_elements = values
                # Set interval
                params.set_param(
                    f"{self.active_axis}Interval", interval, "Mod/Draft/OrthoArrayLinearMode"
                )
                # Set number of elements
                params.set_param(
                    f"{self.active_axis}NumOfElements",
                    num_elements,
                    "Mod/Draft/OrthoArrayLinearMode",
                )

            self.create_object()
            # The internal function already displays messages
            self.finish()

    def validate_input(self, selection, v_x, v_y, v_z, n_x, n_y, n_z):
        """Check that the input is valid.

        Some values may not need to be checked because
        the interface may not allow one to input wrong data.
        """
        if not selection:
            _err(translate("draft", "At least 1 element must be selected"))
            return False

        if n_x < 1 or n_y < 1 or n_z < 1:
            _err(translate("draft", "Number of elements must be at least 1"))
            return False

        # TODO: this should handle multiple objects.
        # Each of the elements of the selection should be tested.
        obj = selection[0]
        if obj.isDerivedFrom("App::FeaturePython"):
            _err(translate("draft", "Selection is not suitable for array"))
            _err(translate("draft", "Object:") + " {0} ({1})".format(obj.Label, obj.TypeId))
            return False

        # we should not ever do this but maybe a sanity check here?
        if self.linear_mode:
            if not (
                self.form.radiobutton_x_axis.isChecked()
                or self.form.radiobutton_y_axis.isChecked()
                or self.form.radiobutton_z_axis.isChecked()
            ):
                _err(translate("draft", "In linear mode, at least 1 axis must be selected"))
                return False

        # The other arguments are not tested but they should be present.
        if v_x and v_y and v_z:
            pass

        self.fuse = self.form.checkbox_fuse.isChecked()
        self.use_link = self.form.checkbox_link.isChecked()
        return True

    def create_object(self):
        """Create the new object.

        At this stage we already tested that the input is correct
        so the necessary attributes are already set.
        Then we proceed with the internal function to create the new object.
        """
        if len(self.selection) == 1:
            sel_obj = self.selection[0]
        else:
            # TODO: this should handle multiple objects.
            # For example, it could take the shapes of all objects,
            # make a compound and then use it as input for the array function.
            sel_obj = self.selection[0]

        # clear this stuff out to be sure we don't preserve something
        # from 3-axis mode
        if self.linear_mode:
            start_val = App.Units.Quantity(100.0, App.Units.Length).Value
            if not self.form.radiobutton_x_axis.isChecked():
                self.n_x = 1
                self.v_x = App.Vector(start_val, 0, 0)
            if not self.form.radiobutton_y_axis.isChecked():
                self.n_y = 1
                self.v_y = App.Vector(0, start_val, 0)
            if not self.form.radiobutton_z_axis.isChecked():
                self.n_z = 1
                self.v_z = App.Vector(0, 0, start_val)

        # This creates the object immediately
        # obj = Draft.make_ortho_array(sel_obj,
        #                              self.v_x, self.v_y, self.v_z,
        #                              self.n_x, self.n_y, self.n_z,
        #                              self.use_link)

        # Instead, we build the commands to execute through the caller
        # of this class, the GuiCommand.
        # This is needed to schedule geometry manipulation
        # that would crash Coin3D if done in the event callback.
        _cmd = "Draft.make_ortho_array"
        _cmd += "("
        _cmd += "App.ActiveDocument." + sel_obj.Name + ", "
        _cmd += "v_x=" + DraftVecUtils.toString(self.v_x) + ", "
        _cmd += "v_y=" + DraftVecUtils.toString(self.v_y) + ", "
        _cmd += "v_z=" + DraftVecUtils.toString(self.v_z) + ", "
        _cmd += "n_x=" + str(self.n_x) + ", "
        _cmd += "n_y=" + str(self.n_y) + ", "
        _cmd += "n_z=" + str(self.n_z) + ", "
        _cmd += "use_link=" + str(self.use_link)
        _cmd += ")"

        Gui.addModule("Draft")

        _cmd_list = [
            "_obj_ = " + _cmd,
            "_obj_.Fuse = " + str(self.fuse),
            "Draft.autogroup(_obj_)",
            "App.ActiveDocument.recompute()",
        ]

        # We commit the command list through the parent command
        self.source_command.commit(translate("draft", "Create Orthogonal Array"), _cmd_list)

    def get_numbers(self):
        """Get the number of elements from the widgets."""
        return (
            self.form.spinbox_n_X.value(),
            self.form.spinbox_n_Y.value(),
            self.form.spinbox_n_Z.value(),
        )

    def get_intervals(self):
        """Get the interval vectors from the widgets."""
        v_x_x_str = self.form.input_X_x.text()
        v_x_y_str = self.form.input_X_y.text()
        v_x_z_str = self.form.input_X_z.text()
        v_x = App.Vector(_quantity(v_x_x_str), _quantity(v_x_y_str), _quantity(v_x_z_str))

        v_y_x_str = self.form.input_Y_x.text()
        v_y_y_str = self.form.input_Y_y.text()
        v_y_z_str = self.form.input_Y_z.text()
        v_y = App.Vector(_quantity(v_y_x_str), _quantity(v_y_y_str), _quantity(v_y_z_str))

        v_z_x_str = self.form.input_Z_x.text()
        v_z_y_str = self.form.input_Z_y.text()
        v_z_z_str = self.form.input_Z_z.text()
        v_z = App.Vector(_quantity(v_z_x_str), _quantity(v_z_y_str), _quantity(v_z_z_str))

        return v_x, v_y, v_z

    def reset_v(self, interval):
        """Reset the interval to zero distance.

        Parameters
        ----------
        interval: str
            Either "X", "Y", "Z", to reset the interval vector
            for that direction.
        """
        if interval == "X":
            self.form.input_X_x.setProperty("rawValue", 100)
            self.form.input_X_y.setProperty("rawValue", 0)
            self.form.input_X_z.setProperty("rawValue", 0)
            self.v_x, self.v_y, self.v_z = self.get_intervals()
        elif interval == "Y":
            self.form.input_Y_x.setProperty("rawValue", 0)
            self.form.input_Y_y.setProperty("rawValue", 100)
            self.form.input_Y_z.setProperty("rawValue", 0)
            self.v_x, self.v_y, self.v_z = self.get_intervals()
        elif interval == "Z":
            self.form.input_Z_x.setProperty("rawValue", 0)
            self.form.input_Z_y.setProperty("rawValue", 0)
            self.form.input_Z_z.setProperty("rawValue", 100)
            self.v_x, self.v_y, self.v_z = self.get_intervals()

    def print_fuse_state(self, fuse):
        """Print the fuse state translated."""
        if fuse:
            state = self.tr_true
        else:
            state = self.tr_false
        _msg(translate("draft", "Fuse:") + " {}".format(state))

    def set_fuse(self):
        """Execute as a callback when the fuse checkbox changes."""
        self.fuse = self.form.checkbox_fuse.isChecked()
        params.set_param("Draft_array_fuse", self.fuse)

    def print_link_state(self, use_link):
        """Print the link state translated."""
        if use_link:
            state = self.tr_true
        else:
            state = self.tr_false
        _msg(translate("draft", "Create link array:") + " {}".format(state))

    def set_link(self):
        """Execute as a callback when the link checkbox changes."""
        self.use_link = self.form.checkbox_link.isChecked()
        params.set_param("Draft_array_Link", self.use_link)

    def print_messages(self):
        """Print messages about the operation."""
        if len(self.selection) == 1:
            sel_obj = self.selection[0]
        else:
            # TODO: this should handle multiple objects.
            # For example, it could take the shapes of all objects,
            # make a compound and then use it as input for the array function.
            sel_obj = self.selection[0]
        _msg(translate("draft", "Object:") + " {}".format(sel_obj.Label))

        _msg(translate("draft", "Number of X elements:") + " {}".format(self.n_x))
        _msg(
            translate("draft", "Interval X:")
            + " ({0}, {1}, {2})".format(self.v_x.x, self.v_x.y, self.v_x.z)
        )
        _msg(translate("draft", "Number of Y elements:") + " {}".format(self.n_y))
        _msg(
            translate("draft", "Interval Y:")
            + " ({0}, {1}, {2})".format(self.v_y.x, self.v_y.y, self.v_y.z)
        )
        _msg(translate("draft", "Number of Z elements:") + " {}".format(self.n_z))
        _msg(
            translate("draft", "Interval Z:")
            + " ({0}, {1}, {2})".format(self.v_z.x, self.v_z.y, self.v_z.z)
        )
        self.print_fuse_state(self.fuse)
        self.print_link_state(self.use_link)

    def reject(self):
        """Execute when clicking the Cancel button or pressing Escape."""
        self.finish()

    def toggle_linear_mode(self):
        """Toggle between Linear mode and Orthogonal mode."""
        self.linear_mode = self.form.button_linear_mode.isChecked()
        self.toggle_axis_radiobuttons(self.linear_mode)
        params.set_param("LinearModeOn", self.linear_mode, "Mod/Draft/OrthoArrayLinearMode")

        if self.linear_mode:
            self.form.button_linear_mode.setText(translate("draft", "Switch to Ortho Mode"))

            # check radiobutton based on current cfg
            self.update_axis_ui()

            # For linear mode we're hiding all group boxes for X-, Y-, Z-Axis and the one
            # with number of elements as we will reparent those spinboxes under newly
            # created group
            self._set_orthomode_groups_visibility(hide=True)
            self._reparent_groups(mode="Linear")

            # Set the appropriate title for the group (we flip it back and forth after changing mode)
            # and show the group
            self.form.group_linearmode.show()
            match self.active_axis:
                case "X":
                    title = translate("draft", "X-Axis")
                case "Y":
                    title = translate("draft", "Y-Axis")
                case "Z":
                    title = translate("draft", "Z-Axis")
            self.form.group_linearmode.setTitle(title)
        else:  # ortho mode
            self.form.button_linear_mode.setText(translate("draft", "Switch to Linear Mode"))

            # For ortho mode we're showing back default groupboxes and we reparent everything
            # back to them, as we reuse spinboxes in both modes
            self._set_orthomode_groups_visibility(hide=False)
            self._reparent_groups(mode="Ortho")

            self.form.group_linearmode.hide()

    def toggle_axis_radiobuttons(self, show):
        """Show or hide the axis radio buttons."""
        for radiobutton in [
            self.form.radiobutton_x_axis,
            self.form.radiobutton_y_axis,
            self.form.radiobutton_z_axis,
        ]:
            radiobutton.setVisible(show)

    def set_active_axis(self, axis):
        """Set the active axis when a radio button is changed."""
        if self.linear_mode:
            # get current radiobutton that was supposed to have state changed
            radiobutton = getattr(self.form, f"radiobutton_{axis.lower()}_axis")

            # If we're checking a different radio button than the current active axis
            if radiobutton.isChecked() and axis != self.active_axis:
                self.active_axis = axis
                params.set_param("AxisSelected", self.active_axis, "Mod/Draft/OrthoArrayLinearMode")
                self._setup_linear_mode_layout()
                match self.active_axis:
                    case "X":
                        title = translate("draft", "X-Axis")
                    case "Y":
                        title = translate("draft", "Y-Axis")
                    case "Z":
                        title = translate("draft", "Z-Axis")
                self.form.group_linearmode.setTitle(title)

    def update_axis_ui(self):
        """Update the UI to reflect the current axis selection."""
        # Make sure only one axis is selected
        self.form.radiobutton_x_axis.setChecked(self.active_axis == "X")
        self.form.radiobutton_y_axis.setChecked(self.active_axis == "Y")
        self.form.radiobutton_z_axis.setChecked(self.active_axis == "Z")

    def _get_axis_widgets(self, axis):
        """Get all widgets for a specific axis."""
        return {
            "spinbox_elements": getattr(self.form, f"spinbox_n_{axis}", None),
            "spinbox_interval": getattr(self.form, f"input_{axis}_{axis.lower()}", None),
            "button_reset": getattr(self.form, f"button_reset_{axis}", None),
        }

    def _clear_linear_mode_layout(self):
        """Clear all widgets from the linear mode layout."""
        group_layout = self.form.group_linearmode.layout()

        # Remove all items from the layout
        while group_layout.count():
            item = group_layout.takeAt(0)
            if item.widget():
                item.widget().setParent(None)

    def _setup_linear_mode_layout(self):
        """Set up the Linear mode layout with widgets for the active axis."""
        # Clear existing layout first
        self._clear_linear_mode_layout()

        group_layout = self.form.group_linearmode.layout()

        # Create labels
        label_elements = QtWidgets.QLabel(translate("draft", "Number of elements"))
        label_interval = QtWidgets.QLabel(translate("draft", "Interval"))

        # Get widgets for active axis
        widgets = self._get_axis_widgets(self.active_axis)

        # Add widgets to layout
        widget_pairs = [
            (label_elements, widgets["spinbox_elements"]),
            (label_interval, widgets["spinbox_interval"]),
        ]

        for row_index, (label, widget) in enumerate(widget_pairs):
            label.setParent(self.form.group_linearmode)
            widget.setParent(self.form.group_linearmode)
            group_layout.addWidget(label, row_index, 0)
            group_layout.addWidget(widget, row_index, 1)

        # Add reset button spanning both columns
        widgets["button_reset"].setParent(self.form.group_linearmode)
        group_layout.addWidget(widgets["button_reset"], 2, 0, 1, 2)

    def _restore_axis_to_ortho_layout(self, axis, row_index):
        """Restore widgets for a specific axis back to their original Ortho layout positions."""
        widgets = self._get_axis_widgets(axis)

        # Restore spinbox elements to grid layout
        grid_number_layout = self.form.findChild(QtWidgets.QGridLayout, "grid_number")
        grid_number_layout.addWidget(widgets["spinbox_elements"], row_index, 1)

        # Restore interval input to its axis-specific inner grid layout
        inner_grid_layout = self.form.findChild(QtWidgets.QGridLayout, f"grid_{axis}")
        inner_grid_layout.addWidget(widgets["spinbox_interval"], row_index, 1)

        # Restore reset button to its axis group
        group_box = self.form.findChild(QtWidgets.QGroupBox, f"group_{axis}")
        group_axis_layout = group_box.layout()
        group_axis_layout.addWidget(widgets["button_reset"], 1, 0)

    def _setup_ortho_mode_layout(self):
        """Restore all widgets back to their original Ortho mode layout positions."""
        for row_index, axis in enumerate(["X", "Y", "Z"]):
            self._restore_axis_to_ortho_layout(axis, row_index)

    def _reparent_groups(self, mode="Linear"):
        """Reparent widgets between Linear and Ortho mode layouts."""
        if mode == "Linear":
            self._setup_linear_mode_layout()
        else:
            self._setup_ortho_mode_layout()

    def _set_orthomode_groups_visibility(self, hide=True):
        """Change visibility of ortho mode groups"""
        for axis in ["X", "Y", "Z"]:
            group_name = f"group_{axis}"
            group_box = self.form.findChild(QtWidgets.QGroupBox, group_name)
            if hide:
                group_box.hide()
                self.form.group_copies.hide()
            else:
                group_box.show()
                self.form.group_copies.show()

    def finish(self):
        """Finish the command, after accept or reject.

        It finally calls the parent class to execute
        the delayed functions, and perform cleanup.
        """
        # App.ActiveDocument.commitTransaction()
        if Gui.ActiveDocument is not None:
            Gui.ActiveDocument.resetEdit()
        # Runs the parent command to complete the call
        self.source_command.completed()


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2024 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""BIM Panel-related Arch_"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate

PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")


class Arch_Profile:
    """The FreeCAD Arch_Profile command definition"""

    def GetResources(self):

        return {
            "Pixmap": "Arch_Profile",
            "MenuText": QT_TRANSLATE_NOOP("Arch_Profile", "Profile"),
            "Accel": "P, F",
            "ToolTip": QT_TRANSLATE_NOOP("Arch_Profile", "Creates a profile"),
        }

    def IsActive(self):

        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):

        import ArchProfile

        FreeCAD.activeDraftCommand = self  # register as a Draft command for auto grid on/off
        self.doc = FreeCAD.ActiveDocument
        self.Profile = None
        self.Categories = []
        self.Presets = ArchProfile.readPresets()
        for pre in self.Presets:
            if pre[1] not in self.Categories:
                self.Categories.append(pre[1])
        FreeCADGui.Snapper.getPoint(
            callback=self.getPoint,
            extradlg=[self.taskbox()],
            title=translate("Arch", "Create profile"),
        )

    def taskbox(self):
        "sets up a taskbox widget"

        from PySide import QtGui
        from draftutils import params

        w = QtGui.QWidget()
        ui = FreeCADGui.UiLoader()
        w.setWindowTitle(translate("Arch", "Profile settings"))
        grid = QtGui.QGridLayout(w)

        # categories box
        labelc = QtGui.QLabel(translate("Arch", "Category"))
        self.vCategory = QtGui.QComboBox()
        self.vCategory.addItems(self.Categories)
        grid.addWidget(labelc, 1, 0, 1, 1)
        grid.addWidget(self.vCategory, 1, 1, 1, 1)

        # presets box
        labelp = QtGui.QLabel(translate("Arch", "Preset"))
        self.vPresets = QtGui.QComboBox()
        self.pSelect = [None]
        fpresets = [" "]
        self.vPresets.addItems(fpresets)
        grid.addWidget(labelp, 2, 0, 1, 1)
        grid.addWidget(self.vPresets, 2, 1, 1, 1)

        # restore preset
        categoryIdx = -1
        presetIdx = -1
        stored = params.get_param_arch("ProfilePreset")
        if stored and ";" in stored and len(stored.split(";")) >= 3:
            stored = stored.split(";")
            if stored[1] in self.Categories:
                categoryIdx = self.Categories.index(stored[1])
                self.setCategory(categoryIdx)
                self.vCategory.setCurrentIndex(categoryIdx)
                ps = [p[2] for p in self.pSelect]
                if stored[2] in ps:
                    presetIdx = ps.index(stored[2])
                    self.setPreset(presetIdx)
                    self.vPresets.setCurrentIndex(presetIdx)
        if categoryIdx == -1:
            self.setCategory(0)
            self.vCategory.setCurrentIndex(0)
        if presetIdx == -1:
            self.setPreset(0)
            self.vPresets.setCurrentIndex(0)

        # connect slots
        self.vCategory.currentIndexChanged.connect(self.setCategory)
        self.vPresets.currentIndexChanged.connect(self.setPreset)

        return w

    def getPoint(self, point=None, obj=None):
        "this function is called by the snapper when it has a 3D point"

        FreeCAD.activeDraftCommand = None
        FreeCADGui.Snapper.off()
        if not point:
            return
        if not self.Profile:
            return
        pt = "FreeCAD.Vector(" + str(point.x) + "," + str(point.y) + "," + str(point.z) + ")"
        self.doc.openTransaction(translate("Arch", "Create Profile"))
        FreeCADGui.addModule("Arch")
        FreeCADGui.doCommand("p = Arch.makeProfile(" + str(self.Profile) + ")")
        FreeCADGui.addModule("WorkingPlane")
        FreeCADGui.doCommand("p.Placement = WorkingPlane.get_working_plane().get_placement()")
        FreeCADGui.doCommand("p.Placement.Base = " + pt)
        FreeCADGui.addModule("Draft")
        FreeCADGui.doCommand("Draft.autogroup(p)")
        self.doc.commitTransaction()
        self.doc.recompute()

    def setCategory(self, i):

        from draftutils import params

        self.vPresets.clear()
        self.pSelect = [p for p in self.Presets if p[1] == self.Categories[i]]
        fpresets = []
        for p in self.pSelect:
            f = FreeCAD.Units.Quantity(p[4], FreeCAD.Units.Length).getUserPreferred()
            d = params.get_param("Decimals", path="Units")
            s1 = str(round(p[4] / f[1], d))
            s2 = str(round(p[5] / f[1], d))
            s3 = str(f[2])
            fpresets.append(p[2] + " (" + s1 + "x" + s2 + s3 + ")")
        self.vPresets.addItems(fpresets)
        self.setPreset(0)

    def setPreset(self, i):

        from draftutils import params

        self.Profile = self.pSelect[i]
        params.set_param_arch("ProfilePreset", ";".join([str(i) for i in self.Profile]))


FreeCADGui.addCommand("Arch_Profile", Arch_Profile())



# SPDX-License-Identifier: LGPL-2.1-or-later

## @package importAirfoilDAT
#  \ingroup DRAFT
#  \brief Airfoil (.dat) file importer
#
# This module provides support for importing airfoil .dat files
"""@package importAirfoilDAT
Airfoil (.dat) file importer

This module provides support for importing airfoil .dat files.
"""
# Check code with
# flake8 --ignore=E226,E266,E401,W503

# ***************************************************************************
# *   Copyright (c) 2010 Heiko Jakob <heiko.jakob@gediegos.de>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD Draft Workbench - Airfoil data importer"
__author__ = "Heiko Jakob <heiko.jakob@gediegos.de>"
__url__ = "https://www.freecad.org"

import re
import os
import FreeCAD
import Draft
import Part
from FreeCAD import Vector
from FreeCAD import Console as FCC
from draftutils.utils import pyopen

if FreeCAD.GuiUp:
    from draftutils.translate import translate
else:

    def translate(context, txt):
        return txt


useDraftWire = True


def open(filename):
    """Open filename and parse.

    Parameters
    ----------
    filename : str
        The path to the filename to be opened.

    Returns
    -------
    App::Document
        The new FreeCAD document object created, with the parsed information.
    """
    docname = os.path.splitext(os.path.basename(filename))[0]
    doc = FreeCAD.newDocument(docname)
    doc.Label = docname
    process(filename)
    doc.recompute()


def insert(filename, docname):
    """Get an active document and parse.

    If no document exists, it is created.

    Parameters
    ----------
    filename : str
        The path to the filename to be opened.
    docname : str
        The name of the active App::Document if one exists, or
        of the new one created.

    Returns
    -------
    App::Document
        The active FreeCAD document, or the document created if none exists,
        with the parsed information.
    """
    groupname = os.path.splitext(os.path.basename(filename))[0]
    try:
        doc = FreeCAD.getDocument(docname)
    except NameError:
        doc = FreeCAD.newDocument(docname)
    obj = process(filename)
    if obj is not None:
        importgroup = doc.addObject("App::DocumentObjectGroup", groupname)
        importgroup.Group = [obj]
    doc.recompute()


def process(filename):
    """Process the filename and create a Draft Wire from the data.

    The common airfoil dat format has many flavors.
    This code should work with almost every dialect.

    Parameters
    ----------
    filename : str
        The path to the filename to be opened.

    Returns
    -------
    Part::Feature or None.
        The created object or None if the file contains fewer
        than 3 points.
    """
    # Regex to identify data rows and throw away unused metadata
    xval = r"(?P<xval>\-?\s*\d*\.*\d*([Ee]\-?\d+)?)"
    yval = r"(?P<yval>\-?\s*\d*\.*\d*([Ee]\-?\d+)?)"
    _regex = r"^\s*" + xval + r"\,?\s*" + yval + r"\s*$"

    regex = re.compile(_regex)
    afile = pyopen(filename, "r")
    # read the airfoil name which is always in the first line
    airfoilname = afile.readline().strip()

    coords = []
    # upside = True
    # last_x = None

    # Collect the data
    for lin in afile:
        curdat = regex.match(lin)
        if curdat is not None and curdat.group("xval") and curdat.group("yval"):
            x = float(curdat.group("xval"))
            y = float(curdat.group("yval"))

            # Some files specify the number of upper and lower points on the 2nd line:
            # "       67.       72."
            # See: http://airfoiltools.com/airfoil
            # This line must be skipped:
            if x < 2 and y < 2:
                # the normal processing
                coords.append(Vector(x, y, 0))

    afile.close()

    if len(coords) < 3:
        FCC.PrintError(translate("ImportAirfoilDAT", "Did not find enough coordinates") + "\n")
        return None

    # sometimes coords are divided in upper an lower side
    # so that x-coordinate begin new from leading or trailing edge
    # check for start coordinates in the middle of list
    if coords[0:-1].count(coords[0]) > 1:
        flippoint = coords.index(coords[0], 1)
        upper = coords[0:flippoint]
        lower = coords[flippoint + 1 :]
        lower.reverse()
        for i in lower:
            upper.append(i)
        coords = upper

    # do we use the parametric Draft Wire?
    if useDraftWire:
        obj = Draft.make_wire(coords, True)
        # obj.label = airfoilname
    else:
        # alternate solution, uses common Part Faces
        lines = []
        first_v = None
        last_v = None
        for v in coords:
            if first_v is None:
                first_v = v

            # Line between v and last_v if they're not equal
            if (last_v is not None) and (last_v != v):
                lines.append(Part.makeLine(last_v, v))

            # The new last_v
            last_v = v

        # close the wire if needed
        if last_v != first_v:
            lines.append(Part.makeLine(last_v, first_v))

        wire = Part.Wire(lines)
        face = Part.Face(wire)
        obj = FreeCAD.ActiveDocument.addObject("Part::Feature", airfoilname)
        obj.Shape = face

    return obj



# SPDX-License-Identifier: LGPL-2.1-or-later

import FreeCAD
import Part

from .CompoundFilter import makeCompoundFilter


def explodeCompound(compound_obj, b_group=None):
    """explodeCompound(compound_obj, b_group = None): creates a bunch of compound filters, to extract every child of a compound into a separate object.
    group: if True, Group is always made. If False, group is never made. If None, group is made if there is more than one child.
    returns: (group_object, list_of_child_objects)"""

    if isinstance(compound_obj, FreeCAD.GeoFeature) and isinstance(
        compound_obj.getPropertyOfGeometry(), Part.Shape
    ):
        sh = compound_obj.getPropertyOfGeometry()
    else:
        raise TypeError("Object must be App.GeoFeature with Part.Shape property")

    n = len(sh.childShapes(False, False))
    if b_group is None:
        b_group = n > 1
    if b_group:
        group = compound_obj.Document.addObject(
            "App::DocumentObjectGroup", "GrExplode_" + compound_obj.Name
        )
        group.Label = "Exploded {obj.Label}".format(obj=compound_obj)
    else:
        group = compound_obj.Document
    features_created = []
    for i in range(0, n):
        cf = makeCompoundFilter(
            "{obj.Name}_child{child_num}".format(obj=compound_obj, child_num=i), group
        )
        cf.Label = "{obj.Label}.{child_num}".format(obj=compound_obj, child_num=i)
        cf.Base = compound_obj
        cf.FilterType = "specific items"
        cf.items = str(i)
        if cf.ViewObject is not None:
            cf.ViewObject.DontUnhideOnDelete = True
        features_created.append(cf)
    return (group, features_created)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 FreeCAD Developers                                 *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to join wires together into a single wire."""
## @package join
# \ingroup draftfunctions
# \brief Provides functions to join wires together into a single wire.

## \addtogroup draftfunctions
# @{
import FreeCAD as App
import DraftVecUtils


def join_wires(wires, joinAttempts=0):
    """join_wires(objects): merges a set of wires where possible, if any of those
    wires have a coincident start and end point"""
    wires = list(wires)
    if joinAttempts > len(wires):
        return
    joinAttempts += 1
    for wire1Index, wire1 in enumerate(wires):
        for wire2Index, wire2 in enumerate(wires):
            if wire2Index <= wire1Index:
                continue
            if join_two_wires(wire1, wire2):
                wires.pop(wire2Index)
                break
    join_wires(wires, joinAttempts)
    return wires


joinWires = join_wires


def join_two_wires(wire1, wire2):
    """join_two_wires(object, object): joins two wires if they share a common
    point as a start or an end.
    """
    wire1AbsPoints = [wire1.Placement.multVec(point) for point in wire1.Points]
    wire2AbsPoints = [wire2.Placement.multVec(point) for point in wire2.Points]
    if (
        DraftVecUtils.equals(wire1AbsPoints[0], wire2AbsPoints[-1])
        and DraftVecUtils.equals(wire1AbsPoints[-1], wire2AbsPoints[0])
    ) or (
        DraftVecUtils.equals(wire1AbsPoints[0], wire2AbsPoints[0])
        and DraftVecUtils.equals(wire1AbsPoints[-1], wire2AbsPoints[-1])
    ):
        wire2AbsPoints.pop()
        wire1.Closed = True
    elif DraftVecUtils.equals(wire1AbsPoints[0], wire2AbsPoints[0]):
        wire1AbsPoints = list(reversed(wire1AbsPoints))
    elif DraftVecUtils.equals(wire1AbsPoints[0], wire2AbsPoints[-1]):
        wire1AbsPoints = list(reversed(wire1AbsPoints))
        wire2AbsPoints = list(reversed(wire2AbsPoints))
    elif DraftVecUtils.equals(wire1AbsPoints[-1], wire2AbsPoints[-1]):
        wire2AbsPoints = list(reversed(wire2AbsPoints))
    elif DraftVecUtils.equals(wire1AbsPoints[-1], wire2AbsPoints[0]):
        pass
    else:
        return False
    wire2AbsPoints.pop(0)
    wire1.Points = [wire1.Placement.inverse().multVec(point) for point in wire1AbsPoints] + [
        wire1.Placement.inverse().multVec(point) for point in wire2AbsPoints
    ]
    App.ActiveDocument.removeObject(wire2.Name)
    return True


joinTwoWires = join_two_wires

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2018 Benjamin Alterauge (ageeye)                        *
# *   Copyright (c) 2020 Carlo Pavan <carlopav@gmail.com>                   *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the object code for the PointArray object."""
## @package pointarray
# \ingroup draftobjects
# \brief Provides the object code for the PointArray object.

from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import DraftVecUtils
import draftutils.utils as utils

from draftutils.messages import _err, _log
from draftutils.translate import translate
from draftobjects.draftlink import DraftLink

## \addtogroup draftobjects
# @{


class PointArray(DraftLink):
    """The Draft Point Array object."""

    def __init__(self, obj):
        super().__init__(obj, "PointArray")

    def attach(self, obj):
        """Set up the properties when the object is attached."""
        self.set_properties(obj)
        super().attach(obj)

    def linkSetup(self, obj):
        """Set up the object as a link object."""
        super().linkSetup(obj)
        obj.configLinkProperty(ElementCount="Count")

    def set_properties(self, obj):
        """Set properties only if they don't exist."""
        properties = obj.PropertiesList

        if "Base" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Base object that will be duplicated")
            obj.addProperty("App::PropertyLink", "Base", "Objects", _tip, locked=True)
            obj.Base = None

        if "PointObject" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property", "Object containing points used to distribute the copies."
            )
            obj.addProperty("App::PropertyLink", "PointObject", "Objects", _tip, locked=True)
            obj.PointObject = None

        if "Fuse" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "Specifies if the copies "
                "should be fused together "
                "if they touch each other (slower)",
            )
            obj.addProperty("App::PropertyBool", "Fuse", "Objects", _tip, locked=True)
            obj.Fuse = False

        if "Count" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "Number of copies in the array.\nThis property is read-only, as the number depends on the points in 'Point Object'.",
            )
            obj.addProperty("App::PropertyInteger", "Count", "Objects", _tip, locked=True)
            obj.Count = 0
            obj.setEditorMode("Count", 1)  # Read only

        if "ExtraPlacement" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "Additional placement, shift and rotation, that will be applied to each copy",
            )
            obj.addProperty(
                "App::PropertyPlacement", "ExtraPlacement", "Objects", _tip, locked=True
            )
            obj.ExtraPlacement = App.Placement()

        if self.use_link and "ExpandArray" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property", "Show the individual array elements (only for Link arrays)"
            )
            obj.addProperty("App::PropertyBool", "ExpandArray", "Objects", _tip, locked=True)
            obj.setPropertyStatus("Shape", "Transient")

        if not self.use_link:
            if "PlacementList" not in properties:
                _tip = QT_TRANSLATE_NOOP("App::Property", "The placement for each array element")
                obj.addProperty(
                    "App::PropertyPlacementList", "PlacementList", "Objects", _tip, locked=True
                )
                obj.PlacementList = []

    def execute(self, obj):
        """Run when the object is created or recomputed."""
        if self.props_changed_placement_only(obj) or not obj.Base or not obj.PointObject:
            self.props_changed_clear()
            return

        pt_list = get_point_list(obj.PointObject)
        obj.Count = len(pt_list)
        pls = build_placements(obj.Base, pt_list, obj.ExtraPlacement)

        self.buildShape(obj, obj.Placement, pls)
        self.props_changed_clear()
        return not self.use_link

    def onDocumentRestored(self, obj):
        super().onDocumentRestored(obj)
        # Fuse property was added in v1.0 and PlacementList property was added
        # for non-link arrays in v1.1, obj should be OK if both are present:
        if hasattr(obj, "Fuse") and hasattr(obj, "PlacementList"):
            return

        if not hasattr(obj, "ExtraPlacement"):
            _log("v0.19, " + obj.Name + ", added 'ExtraPlacement' property")
        if hasattr(obj, "PointList"):
            _log("v0.19, " + obj.Name + ", migrated 'PointList' property to 'PointObject'")
        if not hasattr(obj, "Fuse"):
            _log("v1.0, " + obj.Name + ", added 'Fuse' property")
        if not hasattr(obj, "PlacementList"):
            _log("v1.1, " + obj.Name + ", added hidden property 'PlacementList'")

        self.set_properties(obj)
        if hasattr(obj, "PointList"):
            obj.PointObject = obj.PointList
            obj.removeProperty("PointList")
        self.execute(obj)  # Required to update PlacementList.


def remove_equal_vecs(vec_list):
    """Remove equal vectors from a list.

    Parameters
    ----------
    vec_list: list of App.Vectors

    Returns
    -------
    list of App.Vectors
    """
    res_list = []
    for vec in vec_list:
        for res in res_list:
            if DraftVecUtils.equals(vec, res):
                break
        else:
            res_list.append(vec)
    return res_list


def get_point_list(point_object):
    """Extract a list of points from a point object.

    Parameters
    ----------
    point_object: Part::Feature, Sketcher::SketchObject, Mesh::Feature
                  or Points::FeatureCustom
        The object must have vertices and/or points.

    Returns
    -------
    list of App.Vectors
    """
    if hasattr(point_object, "Shape") and hasattr(point_object.Shape, "Vertexes"):
        pt_list = [v.Point for v in point_object.Shape.Vertexes]
        # For compatibility with previous versions: add all points from sketch (including construction geometry):
        if hasattr(point_object, "Geometry"):
            place = point_object.Placement
            for geo in point_object.Geometry:
                if geo.TypeId == "Part::GeomPoint":
                    pt_list.append(place.multVec(App.Vector(geo.X, geo.Y, geo.Z)))
    elif hasattr(point_object, "Mesh"):
        pt_list = [p.Vector for p in point_object.Mesh.Points]
    elif hasattr(point_object, "Points"):
        pt_list = point_object.Points.Points
    else:
        return []

    return remove_equal_vecs(pt_list)


def build_placements(base_object, pt_list=None, placement=App.Placement()):
    """Build a placements from the base object and list of points.

    Returns
    -------
    list of App.Placements
    """
    if not pt_list:
        _err(
            translate(
                "Draft",
                "Point object does not have a discrete point, " "it cannot be used for an array",
            )
        )
        return []

    pls = list()

    for point in pt_list:
        new_pla = base_object.Placement.copy()
        original_rotation = new_pla.Rotation

        # Reset the position of the copy, and combine the original rotation
        # with the provided rotation. Two rotations (quaternions)
        # are combined by multiplying them.
        new_pla.Base = placement.Base
        new_pla.Rotation = original_rotation * placement.Rotation
        new_pla.translate(point)

        pls.append(new_pla)

    return pls


# Alias for compatibility with v0.18 and earlier
_PointArray = PointArray

## @}



# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM calculix write inpfile materials"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"


import FreeCAD
from femtools import constants


def write_femelement_material(f, ccxwriter):

    # floats read from ccx should use {:.13G}, see comment in writer module
    # see unit comment in writer module

    def is_density_needed():
        if ccxwriter.analysis_type == "frequency":
            return True
        if (
            ccxwriter.analysis_type == "thermomech"
            and not ccxwriter.solver_obj.ThermoMechSteadyState
        ):
            return True
        if ccxwriter.member.cons_centrif:
            return True
        if ccxwriter.member.cons_selfweight:
            return True
        return False

    f.write("\n** Physical constants for SI(mm) unit system with Kelvins\n")
    f.write("*PHYSICAL CONSTANTS, ABSOLUTE ZERO=0, STEFAN BOLTZMANN=5.670374419e-11\n")

    f.write("\n{}\n".format(59 * "*"))
    f.write("** Materials\n")
    f.write("** see information about units at file end\n")
    for femobj in ccxwriter.member.mats_linear:
        # femobj --> dict, FreeCAD document object is femobj["Object"]
        mat_obj = femobj["Object"]
        mat_info_name = mat_obj.Material["Name"]
        mat_name = mat_obj.Name
        mat_label = mat_obj.Label

        # get material properties of solid material, Currently in SI units: M/kg/s/Kelvin
        if mat_obj.Category == "Solid":
            YM = FreeCAD.Units.Quantity(mat_obj.Material["YoungsModulus"])
            YM_in_MPa = YM.getValueAs("MPa").Value
            PR = float(mat_obj.Material["PoissonRatio"])
        if is_density_needed() is True:
            density = FreeCAD.Units.Quantity(mat_obj.Material["Density"])
            density_in_tonne_per_mm3 = density.getValueAs("t/mm^3").Value
        if ccxwriter.analysis_type == "thermomech":
            TC = FreeCAD.Units.Quantity(mat_obj.Material["ThermalConductivity"])
            # SvdW: Add factor to force units to results base units
            # of t/mm/s/K - W/m/K results in no factor needed
            TC_in_WmK = TC.getValueAs("W/m/K").Value
            SH = FreeCAD.Units.Quantity(mat_obj.Material["SpecificHeat"])
            # SvdW: Add factor to force units to results base units of t/mm/s/K
            # FIXME: why not get it directly in the units needed ?
            SH_in_JkgK = SH.getValueAs("J/kg/K").Value * 1e06
            if mat_obj.Category == "Solid":
                TEC = FreeCAD.Units.Quantity(mat_obj.Material["ThermalExpansionCoefficient"])
                TEC_in_mmK = TEC.getValueAs("mm/mm/K").Value
                if "ThermalExpansionReferenceTemperature" in mat_obj.Material:
                    RT = FreeCAD.Units.Quantity(
                        mat_obj.Material["ThermalExpansionReferenceTemperature"]
                    )
                else:
                    RT = FreeCAD.Units.Quantity("0 K")
                RT_in_K = RT.getValueAs("K").Value
            elif mat_obj.Category == "Fluid":
                KV = FreeCAD.Units.Quantity(mat_obj.Material["KinematicViscosity"])
                KV_in_mm2s = KV.getValueAs("mm^2/s").Value
                DV_in_tmms = KV_in_mm2s * density_in_tonne_per_mm3
        if ccxwriter.analysis_type == "static":
            if mat_obj.Category == "Solid":
                if "ThermalExpansionCoefficient" in mat_obj.Material:
                    TEC = FreeCAD.Units.Quantity(mat_obj.Material["ThermalExpansionCoefficient"])
                    TEC_in_mmK = TEC.getValueAs("mm/mm/K").Value
                else:
                    TEC_in_mmK = 0.0
                if "ThermalExpansionReferenceTemperature" in mat_obj.Material:
                    RT = FreeCAD.Units.Quantity(
                        mat_obj.Material["ThermalExpansionReferenceTemperature"]
                    )
                else:
                    RT = FreeCAD.Units.Quantity("0 K")
                RT_in_K = RT.getValueAs("K").Value
        if (
            ccxwriter.analysis_type == "electromagnetic"
            and ccxwriter.solver_obj.ElectromagneticMode == "electrostatic"
        ):
            rel_perm = FreeCAD.Units.Quantity(mat_obj.Material["RelativePermittivity"]).Value
            vacuum_perm = FreeCAD.Units.Quantity(constants.vacuum_permittivity())
            abs_perm = vacuum_perm.getValueAs("C/(mV*mm)").Value * rel_perm
        # write material properties
        f.write(f"** FreeCAD material name: {mat_info_name}\n")
        f.write(f"** {mat_label}\n")
        f.write(f"*MATERIAL, NAME={mat_name}\n")
        if mat_obj.Category == "Solid":
            f.write("*ELASTIC\n")
            f.write(f"{YM_in_MPa:.13G},{PR:.13G}\n")
        if is_density_needed() is True:
            f.write("*DENSITY\n")
            f.write(f"{density_in_tonne_per_mm3:.13G}\n")
        if ccxwriter.analysis_type == "thermomech":
            if mat_obj.Category == "Solid":
                f.write("*CONDUCTIVITY\n")
                f.write(f"{TC_in_WmK:.13G}\n")
                f.write(f"*EXPANSION, ZERO={RT_in_K:.13G}\n")
                f.write(f"{TEC_in_mmK:.13G}\n")
                f.write("*SPECIFIC HEAT\n")
                f.write(f"{SH_in_JkgK:.13G}\n")
            elif mat_obj.Category == "Fluid":
                f.write("*FLUID CONSTANTS\n")
                f.write(f"{SH_in_JkgK:.13G},{DV_in_tmms:.13G}\n")
        if ccxwriter.analysis_type == "static":
            if mat_obj.Category == "Solid":
                f.write(f"*EXPANSION, ZERO={RT_in_K:.13G}\n")
                f.write(f"{TEC_in_mmK:.13G}\n")
        if (
            ccxwriter.analysis_type == "electromagnetic"
            and ccxwriter.solver_obj.ElectromagneticMode == "electrostatic"
        ):
            f.write("*CONDUCTIVITY\n")
            f.write(f"{abs_perm:.13G}\n")

        # nonlinear material properties
        if ccxwriter.solver_obj.MaterialNonlinearity == "nonlinear":

            for nlfemobj in ccxwriter.member.mats_nonlinear:
                # femobj --> dict, FreeCAD document object is nlfemobj["Object"]
                nl_mat_obj = nlfemobj["Object"]
                if nl_mat_obj.LinearBaseMaterial == mat_obj:
                    if nl_mat_obj.MaterialModelNonlinearity == "isotropic hardening":
                        f.write("*PLASTIC\n")
                    else:
                        f.write("*PLASTIC, HARDENING=KINEMATIC\n")
                    for yield_point in nl_mat_obj.YieldPoints:
                        f.write(f"{yield_point}\n")
                f.write("\n")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2015 Dan Falck <ddfalck@gmail.com>                      *
# *   Copyright (c) 2021 Schildkroet                                        *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENTE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""Selection gates and observers to control selectability while building Path operations"""

import FreeCAD
import FreeCADGui
import Path
import Path.Base.Drillable as Drillable
import math

Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
Path.Log.trackModule(Path.Log.thisModule())


class PathBaseGate(object):
    pass


class EGate(PathBaseGate):
    def allow(self, doc, obj, sub):
        return sub and sub[0:4] == "Edge"


class MESHGate(PathBaseGate):
    def allow(self, doc, obj, sub):
        return obj.TypeId[0:4] == "Mesh"


class VCARVEGate:
    def allow(self, doc, obj, sub):
        try:
            shape = obj.Shape
        except Exception:
            return False

        if math.fabs(shape.Volume) < 1e-9 and len(shape.Wires) > 0:
            return True

        if shape.ShapeType == "Face":
            return True

        elif shape.ShapeType == "Solid":
            if sub and sub[0:4] == "Face":
                return True

        elif shape.ShapeType == "Compound":
            if sub and sub[0:4] == "Face":
                return True

        if sub:
            subShape = shape.getElement(sub)
            if subShape.ShapeType == "Edge":
                return False

        return False


class ENGRAVEGate(PathBaseGate):
    def allow(self, doc, obj, sub):
        try:
            shape = obj.Shape
        except Exception:
            return False

        if math.fabs(shape.Volume) < 1e-9 and len(shape.Wires) > 0:
            return True

        if shape.ShapeType == "Edge":
            return True

        if sub:
            subShape = shape.getElement(sub)
            if subShape.ShapeType == "Edge":
                return True

        return False


class CHAMFERGate(PathBaseGate):
    def allow(self, doc, obj, sub):
        try:
            shape = obj.Shape
        except Exception:
            return False

        if math.fabs(shape.Volume) < 1e-9 and len(shape.Wires) > 0:
            return True

        if "Edge" == shape.ShapeType or "Face" == shape.ShapeType:
            return True

        if sub:
            subShape = shape.getElement(sub)
            if subShape.ShapeType == "Edge":
                return True
            elif subShape.ShapeType == "Face":
                return True

        return False


class DRILLGate(PathBaseGate):
    def allow(self, doc, obj, sub):
        Path.Log.debug("obj: {} sub: {}".format(obj, sub))
        if not hasattr(obj, "Shape"):
            return False
        shape = obj.Shape
        subobj = shape.getElement(sub)
        if subobj.ShapeType not in ["Edge", "Face"]:
            return False
        return Drillable.isDrillable(shape, subobj, vector=None, allowPartial=True)


class TAPGate(PathBaseGate):
    def allow(self, doc, obj, sub):
        Path.Log.debug("obj: {} sub: {}".format(obj, sub))
        if not hasattr(obj, "Shape"):
            return False
        shape = obj.Shape
        subobj = shape.getElement(sub)
        if subobj.ShapeType not in ["Edge", "Face"]:
            return False
        return Drillable.isDrillable(shape, subobj, vector=None)


class FACEGate(PathBaseGate):
    def allow(self, doc, obj, sub):
        isFace = False

        try:
            obj = obj.Shape
        except Exception:
            return False

        if obj.ShapeType == "Compound":
            if sub and sub[0:4] == "Face":
                isFace = True

        elif obj.ShapeType == "Face":  # 3D Face, not flat, planar?
            isFace = True

        elif obj.ShapeType == "Solid":
            if sub and sub[0:4] == "Face":
                isFace = True

        return isFace


class PROFILEGate(PathBaseGate):
    def allow(self, doc, obj, sub):
        if sub and sub[0:4] == "Edge":
            return True

        try:
            obj = obj.Shape
        except Exception:
            return False

        if obj.ShapeType == "Compound":
            if sub and sub[0:4] == "Face":
                return True

        elif obj.ShapeType == "Face":
            return True

        elif obj.ShapeType == "Solid":
            if sub and sub[0:4] == "Face":
                return True

        elif obj.ShapeType == "Wire":
            return True

        return False


class POCKETGate(PathBaseGate):
    def allow(self, doc, obj, sub):

        pocketable = False
        try:
            obj = obj.Shape
        except Exception:
            return False

        if obj.ShapeType == "Edge":
            pocketable = False

        elif obj.ShapeType == "Face":
            pocketable = True

        elif obj.ShapeType == "Solid":
            if sub and sub[0:4] == "Face":
                pocketable = True

        elif obj.ShapeType == "Compound":
            if sub and sub[0:4] == "Face":
                pocketable = True

        return pocketable


class ADAPTIVEGate(PathBaseGate):
    def allow(self, doc, obj, sub):

        adaptive = True
        try:
            obj = obj.Shape
        except Exception:
            return False

        return adaptive


class CONTOURGate(PathBaseGate):
    def allow(self, doc, obj, sub):
        pass


class PROBEGate:
    def allow(self, doc, obj, sub):
        pass


class TURNGate(PathBaseGate):
    def allow(self, doc, obj, sub):
        Path.Log.debug("obj: {} sub: {}".format(obj, sub))
        if hasattr(obj, "Shape") and sub:
            shape = obj.Shape
            subobj = shape.getElement(sub)
            return Drillable.isDrillable(shape, subobj, vector=None)
        else:
            return False


class ALLGate(PathBaseGate):
    def allow(self, doc, obj, sub):
        if sub and sub[0:6] == "Vertex":
            return True
        if sub and sub[0:4] == "Edge":
            return True
        if sub and sub[0:4] == "Face":
            return True
        return False


def contourselect():
    FreeCADGui.Selection.addSelectionGate(CONTOURGate())
    if not Path.Preferences.suppressSelectionModeWarning():
        FreeCAD.Console.PrintWarning("Contour Select Mode\n")


def eselect():
    FreeCADGui.Selection.addSelectionGate(EGate())
    if not Path.Preferences.suppressSelectionModeWarning():
        FreeCAD.Console.PrintWarning("Edge Select Mode\n")


def drillselect():
    FreeCADGui.Selection.addSelectionGate(DRILLGate())
    if not Path.Preferences.suppressSelectionModeWarning():
        FreeCAD.Console.PrintWarning("Drilling Select Mode\n")


def tapselect():
    FreeCADGui.Selection.addSelectionGate(TAPGate())
    if not Path.Preferences.suppressSelectionModeWarning():
        FreeCAD.Console.PrintWarning("Tapping Select Mode\n")


def engraveselect():
    FreeCADGui.Selection.addSelectionGate(ENGRAVEGate())
    if not Path.Preferences.suppressSelectionModeWarning():
        FreeCAD.Console.PrintWarning("Engraving Select Mode\n")


def fselect():
    FreeCADGui.Selection.addSelectionGate(FACEGate())  # Was PROFILEGate()
    if not Path.Preferences.suppressSelectionModeWarning():
        FreeCAD.Console.PrintWarning("Profiling Select Mode\n")


def chamferselect():
    FreeCADGui.Selection.addSelectionGate(CHAMFERGate())
    if not Path.Preferences.suppressSelectionModeWarning():
        FreeCAD.Console.PrintWarning("Deburr Select Mode\n")


def profileselect():
    FreeCADGui.Selection.addSelectionGate(PROFILEGate())
    if not Path.Preferences.suppressSelectionModeWarning():
        FreeCAD.Console.PrintWarning("Profiling Select Mode\n")


def pocketselect():
    FreeCADGui.Selection.addSelectionGate(POCKETGate())
    if not Path.Preferences.suppressSelectionModeWarning():
        FreeCAD.Console.PrintWarning("Pocketing Select Mode\n")


def adaptiveselect():
    FreeCADGui.Selection.addSelectionGate(ADAPTIVEGate())
    if not Path.Preferences.suppressSelectionModeWarning():
        FreeCAD.Console.PrintWarning("Adaptive Select Mode\n")


def slotselect():
    FreeCADGui.Selection.addSelectionGate(ALLGate())
    if not Path.Preferences.suppressSelectionModeWarning():
        FreeCAD.Console.PrintWarning("Slot Cutter Select Mode\n")


def surfaceselect():
    gate = False
    if MESHGate() or FACEGate():
        gate = True
    FreeCADGui.Selection.addSelectionGate(gate)
    if not Path.Preferences.suppressSelectionModeWarning():
        FreeCAD.Console.PrintWarning("Surfacing Select Mode\n")


def vcarveselect():
    FreeCADGui.Selection.addSelectionGate(VCARVEGate())
    if not Path.Preferences.suppressSelectionModeWarning():
        FreeCAD.Console.PrintWarning("Vcarve Select Mode\n")


def probeselect():
    FreeCADGui.Selection.addSelectionGate(PROBEGate())
    if not Path.Preferences.suppressSelectionModeWarning():
        FreeCAD.Console.PrintWarning("Probe Select Mode\n")


def customselect():
    if not Path.Preferences.suppressSelectionModeWarning():
        FreeCAD.Console.PrintWarning("Custom Select Mode\n")


def turnselect():
    FreeCADGui.Selection.addSelectionGate(TURNGate())
    if not Path.Preferences.suppressSelectionModeWarning():
        FreeCAD.Console.PrintWarning("Turning Select Mode\n")


def select(op):
    opsel = {}
    opsel["Contour"] = contourselect  # deprecated
    opsel["Deburr"] = chamferselect
    opsel["Drilling"] = drillselect
    opsel["Tapping"] = tapselect
    opsel["Engrave"] = engraveselect
    opsel["Helix"] = drillselect
    opsel["MillFace"] = pocketselect
    opsel["Pocket"] = pocketselect
    opsel["Pocket 3D"] = pocketselect
    opsel["Pocket3D"] = pocketselect  # deprecated
    opsel["Pocket Shape"] = pocketselect
    opsel["Profile Edges"] = eselect  # deprecated
    opsel["Profile Faces"] = fselect  # deprecated
    opsel["Profile"] = profileselect
    opsel["Slot"] = slotselect
    opsel["Surface"] = surfaceselect
    opsel["Waterline"] = surfaceselect
    opsel["Adaptive"] = adaptiveselect
    opsel["Vcarve"] = vcarveselect
    opsel["Probe"] = probeselect
    opsel["Custom"] = customselect
    opsel["ThreadMilling"] = drillselect
    opsel["TurnFace"] = turnselect
    opsel["TurnProfile"] = turnselect
    opsel["TurnPartoff"] = turnselect
    opsel["TurnRough"] = turnselect
    return opsel[op]


def clear():
    FreeCADGui.Selection.removeSelectionGate()
    if not Path.Preferences.suppressSelectionModeWarning():
        FreeCAD.Console.PrintWarning("Free Select\n")



# ***************************************************************************
# *   Copyright (c) 2006 Werner Mayer <werner.wm.mayer@gmx.de>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************/

# Workbench test module

import FreeCAD, FreeCADGui, os, unittest
import tempfile

from PySide import QtWidgets, QtCore
from PySide.QtWidgets import QApplication


class CallableCheckWarning:
    def __call__(self):
        diag = QApplication.activeModalWidget()
        if diag:
            QtCore.QTimer.singleShot(0, diag, QtCore.SLOT("accept()"))


class WorkbenchTestCase(unittest.TestCase):
    def setUp(self):
        self.Active = FreeCADGui.activeWorkbench()
        FreeCAD.Console.PrintLog(FreeCADGui.activeWorkbench().name())

    def testActivate(self):
        wbs = FreeCADGui.listWorkbenches()
        # this gives workbenches a possibility to detect that we're under test environment
        FreeCAD.TestEnvironment = True
        for i in wbs:
            try:
                print("Activate workbench '{}'".format(i))
                cobj = CallableCheckWarning()
                QtCore.QTimer.singleShot(500, cobj)
                if FreeCADGui.activeWorkbench().name() != i:
                    success = FreeCADGui.activateWorkbench(i)
                else:
                    # Cannot test activation of an already-active workbench
                    success = True
                FreeCAD.Console.PrintLog(
                    "Active: " + FreeCADGui.activeWorkbench().name() + " Expected: " + i + "\n"
                )
                self.assertTrue(success, "Test on activating workbench {0} failed".format(i))
            except Exception as e:
                self.fail("Loading of workbench '{0}' failed: {1}".format(i, e))
        del FreeCAD.TestEnvironment

    def testHandler(self):
        import __main__

        class UnitWorkbench(__main__.Workbench):
            MenuText = "Unittest"
            ToolTip = "Unittest"

            def Initialize(self):
                cmds = ["Test_Test"]
                self.appendToolbar("My Unittest", cmds)

            def GetClassName(self):
                return "Gui::PythonWorkbench"

        FreeCADGui.addWorkbench(UnitWorkbench())
        wbs = FreeCADGui.listWorkbenches()
        self.assertTrue("UnitWorkbench" in wbs, "Test on adding workbench handler failed")
        FreeCADGui.activateWorkbench("UnitWorkbench")
        FreeCADGui.updateGui()
        self.assertTrue(
            FreeCADGui.activeWorkbench().name() == "UnitWorkbench",
            "Test on loading workbench 'Unittest' failed",
        )
        FreeCADGui.removeWorkbench("UnitWorkbench")
        wbs = FreeCADGui.listWorkbenches()
        self.assertTrue(not "UnitWorkbench" in wbs, "Test on removing workbench handler failed")

    def testInvalidType(self):
        class MyExtWorkbench(FreeCADGui.Workbench):
            def Initialize(self):
                pass

            def GetClassName(self):
                return "App::Extension"

        FreeCADGui.addWorkbench(MyExtWorkbench())
        with self.assertRaises(TypeError):
            FreeCADGui.activateWorkbench("MyExtWorkbench")
        FreeCADGui.removeWorkbench("MyExtWorkbench")

    def tearDown(self):
        FreeCADGui.activateWorkbench(self.Active.name())
        FreeCAD.Console.PrintLog(self.Active.name())


class CommandTestCase(unittest.TestCase):
    def testPR6889(self):
        # Fixes a crash
        TempPath = tempfile.gettempdir()
        macroName = TempPath + os.sep + "testmacro.py"
        macroFile = open(macroName, "w")
        macroFile.write("print ('Hello, World!')")
        macroFile.close()

        name = FreeCADGui.Command.createCustomCommand(macroName)
        cmd = FreeCADGui.Command.get(name)
        cmd.run()


class TestNavigationStyle(unittest.TestCase):
    def setUp(self):
        self.Doc = FreeCAD.newDocument("CreateTest")

    def testInvalidStyle(self):
        FreeCADGui.getDocument(self.Doc).ActiveView.setNavigationType("App::Extension")
        self.assertNotEqual(
            FreeCADGui.getDocument(self.Doc).ActiveView.getNavigationType(), "App::Extension"
        )

    def tearDown(self):
        FreeCAD.closeDocument("CreateTest")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2019, M G Berberich (berberic2, rynn)                   *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the object code for the Array object.

The `Array` class currently handles three types of arrays,
orthogonal, polar, and circular. In the future, probably they should be
split in separate classes so that they are easier to manage.
"""
## @package array
# \ingroup draftobjects
# \brief Provides the object code for the Array object.

## \addtogroup draftobjects
# @{
import math
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import DraftVecUtils

from draftutils.messages import _log

from draftobjects.draftlink import DraftLink


class Array(DraftLink):
    """The Draft Array object.

    To Do
    -----
    The `Array` class currently handles three types of arrays,
    orthogonal, polar, and circular. In the future, probably they should be
    split in separate classes so that they are easier to manage.
    """

    def __init__(self, obj):
        super(Array, self).__init__(obj, "Array")

    def attach(self, obj):
        """Set up the properties when the object is attached."""
        self.set_properties(obj)
        super(Array, self).attach(obj)

    def onDocumentRestored(self, obj):
        super(Array, self).onDocumentRestored(obj)
        # Count property was added in v0.21 and PlacementList property was added
        # for non-link arrays in v1.1, obj should be OK if both are present:
        if hasattr(obj, "Count") and hasattr(obj, "PlacementList"):
            return

        if not hasattr(obj, "Count"):
            _log("v0.21, " + obj.Name + ", added property 'Count'")
        if not hasattr(obj, "PlacementList"):
            _log("v1.1, " + obj.Name + ", added hidden property 'PlacementList'")

        self.set_general_properties(obj)
        self.execute(obj)  # Required to update Count and/or PlacementList.

    def set_properties(self, obj):
        """Set properties only if they don't exist."""
        self.set_ortho_properties(obj)
        self.set_polar_circular_properties(obj)
        self.set_polar_properties(obj)
        self.set_circular_properties(obj)

        # The ArrayType property (general) must be attached
        # after the other array properties so that onChanged works properly
        self.set_general_properties(obj)
        self.set_link_properties(obj)

    def set_general_properties(self, obj):
        """Set general properties only if they don't exist."""
        properties = obj.PropertiesList

        if "Base" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "The base object that will be duplicated")
            obj.addProperty("App::PropertyLink", "Base", "Objects", _tip, locked=True)
            obj.Base = None

        if "ArrayType" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "The type of array to create.\n"
                "- Ortho: places the copies "
                "in the direction of the global X, "
                "Y, Z axes.\n"
                "- Polar: places the copies along "
                "a circular arc, up to a specified "
                "angle, and with certain orientation "
                "defined by a center and an axis.\n"
                "- Circular: places the copies "
                "in concentric circles "
                "around the base object.",
            )
            obj.addProperty("App::PropertyEnumeration", "ArrayType", "Objects", _tip, locked=True)
            obj.ArrayType = ["ortho", "polar", "circular"]

        if "Fuse" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "Specifies if the copies "
                "should be fused together "
                "if they touch each other (slower)",
            )
            obj.addProperty("App::PropertyBool", "Fuse", "Objects", _tip, locked=True)
            obj.Fuse = False

        if "Count" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "Total number of elements "
                "in the array.\n"
                "This property is read-only, "
                "as the number depends "
                "on the parameters of the array.",
            )
            obj.addProperty("App::PropertyInteger", "Count", "Objects", _tip, locked=True)
            obj.Count = 0
            obj.setEditorMode("Count", 1)  # Read only

        if not self.use_link:
            if "PlacementList" not in properties:
                _tip = QT_TRANSLATE_NOOP("App::Property", "The placement for each array element")
                obj.addProperty(
                    "App::PropertyPlacementList", "PlacementList", "Objects", _tip, locked=True
                )
                obj.PlacementList = []

    def set_ortho_properties(self, obj):
        """Set orthogonal properties only if they don't exist."""
        properties = obj.PropertiesList

        if "NumberX" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Number of copies in X-direction")
            obj.addProperty(
                "App::PropertyInteger", "NumberX", "Orthogonal array", _tip, locked=True
            )
            obj.NumberX = 2

        if "NumberY" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Number of copies in Y-direction")
            obj.addProperty(
                "App::PropertyInteger", "NumberY", "Orthogonal array", _tip, locked=True
            )
            obj.NumberY = 2

        if "NumberZ" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Number of copies in Z-direction")
            obj.addProperty(
                "App::PropertyInteger", "NumberZ", "Orthogonal array", _tip, locked=True
            )
            obj.NumberZ = 1

        if "IntervalX" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property", "Distance and orientation " "of intervals in X-direction"
            )
            obj.addProperty(
                "App::PropertyVectorDistance", "IntervalX", "Orthogonal array", _tip, locked=True
            )
            obj.IntervalX = App.Vector(50, 0, 0)

        if "IntervalY" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property", "Distance and orientation " "of intervals in Y-direction"
            )
            obj.addProperty(
                "App::PropertyVectorDistance", "IntervalY", "Orthogonal array", _tip, locked=True
            )
            obj.IntervalY = App.Vector(0, 50, 0)

        if "IntervalZ" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property", "Distance and orientation " "of intervals in Z-direction"
            )
            obj.addProperty(
                "App::PropertyVectorDistance", "IntervalZ", "Orthogonal array", _tip, locked=True
            )
            obj.IntervalZ = App.Vector(0, 0, 50)

    def set_polar_circular_properties(self, obj):
        """Set general polar and circular properties if they don't exist."""
        properties = obj.PropertiesList

        if "Axis" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "The axis direction around which "
                "the elements in a polar or "
                "a circular array will be created",
            )
            obj.addProperty(
                "App::PropertyVector", "Axis", "Polar/circular array", _tip, locked=True
            )
            obj.Axis = App.Vector(0, 0, 1)

        if "Center" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "Center point for polar and "
                "circular arrays.\n"
                "The 'Axis' passes through this point.",
            )
            obj.addProperty(
                "App::PropertyVectorDistance", "Center", "Polar/circular array", _tip, locked=True
            )
            obj.Center = App.Vector(0, 0, 0)

        # The AxisReference property must be attached after Axis and Center
        # so that onChanged works properly
        if "AxisReference" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "The axis object that overrides "
                "the value of 'Axis' and 'Center', "
                "for example, a datum line.\n"
                "Its placement, position and rotation, "
                "will be used when creating polar "
                "and circular arrays.\n"
                "Leave this property empty "
                "to be able to set 'Axis' and 'Center' "
                "manually.",
            )
            obj.addProperty(
                "App::PropertyLinkGlobal", "AxisReference", "Objects", _tip, locked=True
            )
            obj.AxisReference = None

    def set_polar_properties(self, obj):
        """Set polar properties only if they don't exist."""
        properties = obj.PropertiesList

        if "NumberPolar" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Number of copies in the polar direction")
            obj.addProperty("App::PropertyInteger", "NumberPolar", "Polar array", _tip, locked=True)
            obj.NumberPolar = 5

        if "IntervalAxis" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property", "Distance and orientation " "of intervals in 'Axis' direction"
            )
            obj.addProperty(
                "App::PropertyVectorDistance", "IntervalAxis", "Polar array", _tip, locked=True
            )
            obj.IntervalAxis = App.Vector(0, 0, 0)

        if "Angle" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Angle to cover with copies")
            obj.addProperty("App::PropertyAngle", "Angle", "Polar array", _tip, locked=True)
            obj.Angle = 360

    def set_circular_properties(self, obj):
        """Set circular properties only if they don't exist."""
        properties = obj.PropertiesList

        if "RadialDistance" not in properties:
            _tip = QT_TRANSLATE_NOOP("App::Property", "Distance between concentric circles")
            obj.addProperty(
                "App::PropertyDistance", "RadialDistance", "Circular array", _tip, locked=True
            )
            obj.RadialDistance = 50

        if "TangentialDistance" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property", "Distance between copies " "in the same circle"
            )
            obj.addProperty(
                "App::PropertyDistance", "TangentialDistance", "Circular array", _tip, locked=True
            )
            obj.TangentialDistance = 25

        if "NumberCircles" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "Number of concentric circle. " "The 'Base' object counts as one circle.",
            )
            obj.addProperty(
                "App::PropertyInteger", "NumberCircles", "Circular array", _tip, locked=True
            )
            obj.NumberCircles = 3

        if "Symmetry" not in properties:
            _tip = QT_TRANSLATE_NOOP(
                "App::Property",
                "A parameter that determines "
                "how many symmetry planes "
                "the circular array will have",
            )
            obj.addProperty("App::PropertyInteger", "Symmetry", "Circular array", _tip, locked=True)
            obj.Symmetry = 1

    def set_link_properties(self, obj):
        """Set link properties only if they don't exist."""
        properties = obj.PropertiesList

        if self.use_link:
            if "ExpandArray" not in properties:
                _tip = QT_TRANSLATE_NOOP(
                    "App::Property", "Show the individual array elements " "(only for Link arrays)"
                )
                obj.addProperty("App::PropertyBool", "ExpandArray", "Objects", _tip, locked=True)
                obj.ExpandArray = False

    def linkSetup(self, obj):
        """Set up the object as a link object."""
        super(Array, self).linkSetup(obj)
        obj.configLinkProperty(ElementCount="Count")
        obj.setPropertyStatus("Count", "Hidden")

    def onChanged(self, obj, prop):
        """Execute when a property is changed."""
        super(Array, self).onChanged(obj, prop)
        # print(prop, ": ", getattr(obj, prop))
        self.show_and_hide(obj, prop)

    def show_and_hide(self, obj, prop):
        """Show and hide the properties depending on the touched property."""
        if prop == "AxisReference":
            if obj.AxisReference:
                obj.setEditorMode("Center", 1)
                obj.setEditorMode("Axis", 1)
            else:
                obj.setEditorMode("Center", 0)
                obj.setEditorMode("Axis", 0)

        if prop == "ArrayType":
            if obj.ArrayType == "ortho":
                for pr in ("NumberX", "NumberY", "NumberZ", "IntervalX", "IntervalY", "IntervalZ"):
                    obj.setPropertyStatus(pr, "-Hidden")

                for pr in (
                    "Axis",
                    "Center",
                    "NumberPolar",
                    "Angle",
                    "IntervalAxis",
                    "NumberCircles",
                    "RadialDistance",
                    "TangentialDistance",
                    "Symmetry",
                ):
                    obj.setPropertyStatus(pr, "Hidden")

            if obj.ArrayType == "polar":
                for pr in ("Axis", "Center", "NumberPolar", "Angle", "IntervalAxis"):
                    obj.setPropertyStatus(pr, "-Hidden")

                for pr in (
                    "NumberX",
                    "NumberY",
                    "NumberZ",
                    "IntervalX",
                    "IntervalY",
                    "IntervalZ",
                    "NumberCircles",
                    "RadialDistance",
                    "TangentialDistance",
                    "Symmetry",
                ):
                    obj.setPropertyStatus(pr, "Hidden")

            if obj.ArrayType == "circular":
                for pr in (
                    "Axis",
                    "Center",
                    "NumberCircles",
                    "RadialDistance",
                    "TangentialDistance",
                    "Symmetry",
                ):
                    obj.setPropertyStatus(pr, "-Hidden")

                for pr in (
                    "NumberX",
                    "NumberY",
                    "NumberZ",
                    "IntervalX",
                    "IntervalY",
                    "IntervalZ",
                    "NumberPolar",
                    "Angle",
                    "IntervalAxis",
                ):
                    obj.setPropertyStatus(pr, "Hidden")

    def execute(self, obj):
        """Execute when the object is created or recomputed."""
        if self.props_changed_placement_only(obj) or not obj.Base:
            self.props_changed_clear()
            return

        pl = obj.Placement
        axis = obj.Axis
        center = obj.Center

        if hasattr(obj, "AxisReference") and obj.AxisReference:
            if hasattr(obj.AxisReference, "Placement"):
                reference = obj.AxisReference.Placement
                axis = reference.Rotation * App.Vector(0, 0, 1)
                center = reference.Base
            else:
                _info = (
                    "'AxisReference' has no 'Placement' property. "
                    "Please select a different object to use as "
                    "reference."
                )
                raise TypeError(_info)

        if obj.ArrayType == "ortho":
            pls = rect_placements(
                obj.Base.Placement,
                obj.IntervalX,
                obj.IntervalY,
                obj.IntervalZ,
                obj.NumberX,
                obj.NumberY,
                obj.NumberZ,
            )
        elif obj.ArrayType == "polar":
            av = obj.IntervalAxis if hasattr(obj, "IntervalAxis") else None
            pls = polar_placements(
                obj.Base.Placement, center, obj.Angle.Value, obj.NumberPolar, axis, av
            )
        elif obj.ArrayType == "circular":
            pls = circ_placements(
                obj.Base.Placement,
                obj.RadialDistance,
                obj.TangentialDistance,
                axis,
                center,
                obj.NumberCircles,
                obj.Symmetry,
            )

        self.buildShape(obj, pl, pls)
        self.props_changed_clear()
        return not self.use_link


# Alias for compatibility with v0.18 and earlier
_Array = Array


def rect_placements(base_placement, xvector, yvector, zvector, xnum, ynum, znum):
    """Determine the placements where the rectangular copies will be."""
    pl = base_placement
    placements = [pl.copy()]

    for xcount in range(xnum):
        currentxvector = App.Vector(xvector).multiply(xcount)
        if xcount != 0:
            npl = pl.copy()
            npl.translate(currentxvector)
            placements.append(npl)

        for ycount in range(ynum):
            currentyvector = App.Vector(currentxvector)
            _y_shift = App.Vector(yvector).multiply(ycount)
            currentyvector = currentyvector.add(_y_shift)
            if ycount != 0:
                npl = pl.copy()
                npl.translate(currentyvector)
                placements.append(npl)

            for zcount in range(znum):
                currentzvector = App.Vector(currentyvector)
                _z_shift = App.Vector(zvector).multiply(zcount)
                currentzvector = currentzvector.add(_z_shift)
                if zcount != 0:
                    npl = pl.copy()
                    npl.translate(currentzvector)
                    placements.append(npl)

    return placements


def polar_placements(base_placement, center, angle, number, axis, axisvector):
    """Determine the placements where the polar copies will be."""
    # print("angle ",angle," num ",num)
    placements = [base_placement.copy()]

    if number <= 1:
        return placements

    if angle == 360:
        fraction = float(angle) / number
    else:
        fraction = float(angle) / (number - 1)

    for i in range(number - 1):
        currangle = fraction + (i * fraction)
        npl = base_placement.copy()
        npl.rotate(center, axis, currangle, comp=True)
        if axisvector:
            if not DraftVecUtils.isNull(axisvector):
                npl.translate(App.Vector(axisvector).multiply(i + 1))
        placements.append(npl)

    return placements


def circ_placements(
    base_placement, r_distance, tan_distance, axis, center, circle_number, symmetry
):
    """Determine the placements where the circular copies will be."""
    symmetry = max(1, symmetry)
    lead = (0, 1, 0)

    if axis.x == 0 and axis.z == 0:
        lead = (1, 0, 0)

    direction = axis.cross(App.Vector(lead)).normalize()
    placements = [base_placement.copy()]

    for xcount in range(1, circle_number):
        rc = xcount * r_distance
        trans = App.Vector(direction).multiply(rc)
        c = 2 * rc * math.pi
        n = math.floor(c / tan_distance)
        n = int(math.floor(n / symmetry) * symmetry)
        if n == 0:
            continue

        angle = 360.0 / n
        for ycount in range(0, n):
            npl = base_placement.copy()
            npl.translate(trans)
            npl.rotate(center, axis, ycount * angle, comp=True)
            placements.append(npl)

    return placements


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2021 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************


"""This module contains FreeCAD commands for the Draft workbench"""

import os
import FreeCAD
from draftguitools import gui_base
from draftutils import params
from draftutils.todo import todo
from draftutils.translate import QT_TRANSLATE_NOOP, translate


class Draft_Hatch(gui_base.GuiCommandNeedsSelection):

    def GetResources(self):

        return {
            "Pixmap": "Draft_Hatch",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Hatch", "Hatch"),
            "Accel": "H, A",
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_Hatch", "Creates hatches on the faces of a selected object"
            ),
        }

    def Activated(self):

        import FreeCADGui

        if FreeCADGui.Selection.getSelection():
            task = FreeCADGui.Control.showDialog(
                Draft_Hatch_TaskPanel(FreeCADGui.Selection.getSelection()[0])
            )
            task.setDocumentName(FreeCADGui.ActiveDocument.Document.Name)
            task.setAutoCloseOnDeletedDocument(True)
        else:
            FreeCAD.Console.PrintError(
                translate("Draft", "Choose a base object before using this command") + "\n"
            )


class Draft_Hatch_TaskPanel:

    def __init__(self, baseobj):

        import FreeCADGui
        from PySide import QtCore, QtGui
        import Draft_rc

        self.baseobj = baseobj
        self.form = FreeCADGui.PySideUic.loadUi(":/ui/dialogHatch.ui")
        self.form.setWindowIcon(QtGui.QIcon(":/icons/Draft_Hatch.svg"))
        self.form.File.fileNameChanged.connect(self.onFileChanged)
        self.form.File.setFileName(params.get_param("HatchPatternFile"))
        pat = params.get_param("HatchPatternName")
        if pat in [self.form.Pattern.itemText(i) for i in range(self.form.Pattern.count())]:
            self.form.Pattern.setCurrentText(pat)
        self.form.Scale.setValue(params.get_param("HatchPatternScale"))
        self.form.Rotation.setValue(params.get_param("HatchPatternRotation"))
        self.form.Translate.setChecked(params.get_param("HatchPatternTranslate"))
        todo.delay(self.form.setFocus, None)  # Make sure using Esc works.

    def accept(self):

        import FreeCADGui

        params.set_param("HatchPatternFile", self.form.File.property("fileName"))
        params.set_param("HatchPatternName", self.form.Pattern.currentText())
        params.set_param("HatchPatternScale", self.form.Scale.value())
        params.set_param("HatchPatternRotation", self.form.Rotation.value())
        params.set_param("HatchPatternTranslate", self.form.Translate.isChecked())
        if hasattr(self.baseobj, "File") and hasattr(self.baseobj, "Pattern"):
            # modify existing hatch object
            o = 'FreeCAD.ActiveDocument.getObject("' + self.baseobj.Name + '")'
            FreeCADGui.doCommand(o + '.File="' + self.form.File.property("fileName") + '"')
            FreeCADGui.doCommand(o + '.Pattern="' + self.form.Pattern.currentText() + '"')
            FreeCADGui.doCommand(o + ".Scale=" + str(self.form.Scale.value()))
            FreeCADGui.doCommand(o + ".Rotation=" + str(self.form.Rotation.value()))
            FreeCADGui.doCommand(o + ".Translate=" + str(self.form.Translate.isChecked()))
        else:
            # create new hatch object
            FreeCAD.ActiveDocument.openTransaction("Create Hatch")
            FreeCADGui.addModule("Draft")
            cmd = "Draft.make_hatch("
            cmd += 'baseobject=FreeCAD.ActiveDocument.getObject("' + self.baseobj.Name
            cmd += '"),filename="' + self.form.File.property("fileName")
            cmd += '",pattern="' + self.form.Pattern.currentText()
            cmd += '",scale=' + str(self.form.Scale.value())
            cmd += ",rotation=" + str(self.form.Rotation.value())
            cmd += ",translate=" + str(self.form.Translate.isChecked()) + ")"
            FreeCADGui.doCommand(cmd)
            FreeCAD.ActiveDocument.commitTransaction()
        FreeCADGui.doCommand("FreeCAD.ActiveDocument.recompute()")
        self.reject()

    def reject(self):

        import FreeCADGui

        FreeCADGui.Control.closeDialog()
        FreeCADGui.ActiveDocument.resetEdit()
        FreeCAD.ActiveDocument.recompute()

    def onFileChanged(self, filename):

        if filename[0] == "." and self.baseobj:
            # File path relative to the FreeCAD file directory.
            filename = os.path.join(os.path.dirname(self.baseobj.Document.FileName), filename)
            filename = os.path.abspath(filename)

        pat = self.form.Pattern.currentText()
        self.form.Pattern.clear()
        patterns = self.getPatterns(filename)
        self.form.Pattern.addItems(patterns)
        if pat in patterns:
            self.form.Pattern.setCurrentText(pat)

    def getPatterns(self, filename):
        """returns a list of pattern names found in a PAT file"""
        patterns = []
        if os.path.exists(filename):
            with open(filename) as patfile:
                for line in patfile:
                    if line.startswith("*"):
                        patterns.append(line.split(",")[0][1:])
        return patterns


if FreeCAD.GuiUp:
    import FreeCADGui

    FreeCADGui.addCommand("Draft_Hatch", Draft_Hatch())



#! python
# (c) 2010 Werner Mayer LGPL

"""
An example for a high-level custom feature object to make a so called "radial copy".
"""
__author__ = "Werner Mayer <wmayer@users.sourceforge.net>"

import FreeCAD, FreeCADGui, Part, math
from PySide import QtGui
from FreeCAD import Base


def makeCopy(shape, radius, angle):
    mat = Base.Matrix()
    mat.rotateZ(math.radians(angle))
    step = int(360.0 / angle)
    shape = shape.copy()
    shape.translate((radius, 0, 0))
    comp = shape.copy()
    for i in range(step):
        shape.transformShape(mat)
        comp = comp.fuse(shape)
    return comp


class RadialCopy:
    def __init__(self, obj):
        obj.addProperty("App::PropertyLength", "Radius", "", "Radius", locked=True).Radius = 10.0
        obj.addProperty("App::PropertyLength", "Angle", "", "Angle", locked=True).Angle = 20.0
        obj.addProperty("App::PropertyLink", "Source", "", "Source shape", locked=True).Source = (
            None
        )
        obj.Proxy = self

    #   def onChanged(self, fp, prop):
    #       if prop == "Angle" or prop == "Radius":
    #           self.execute(fp)

    def execute(self, fp):
        shape = fp.Source.Shape
        radius = fp.Radius
        angle = fp.Angle
        fp.Shape = makeCopy(shape, radius, angle)


def makeRadialCopy():
    sel = FreeCADGui.Selection.getSelection()
    try:
        sel = sel[0]
        shape = sel.Shape
        name = sel.Label
    except (IndexError, AttributeError):
        QtGui.QMessageBox.critical(None, "Wrong selection", "Please select a shape object")
        # raise Exception("Nothing selected")
    else:
        doc = sel.Document
        rc = doc.addObject("Part::FeaturePython", "RadialCopy")
        rc.Label = name + "(Radial Copy)"
        RadialCopy(rc)
        rc.Source = sel
        rc.ViewObject.Proxy = 0



# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "Mystran add solver control"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## \addtogroup FEM
#  @{


def add_solver_control(f, model, mystran_writer):

    # write the pyNastran code which will be executed into the file
    f.write(pynas_code)

    # print(model.get_bdf_stats())
    exec(pynas_code)
    # print(model.get_bdf_stats())

    return model


pynas_code = """
# executive control
model.sol = 101


# params cards
model.add_param(key="POST", values=-1)
# model.add_param(key="PRTMAXIM", values="YES")  # not recognized by Mystran


# case control
from pyNastran.bdf.bdf import CaseControlDeck
cc = CaseControlDeck([
    "ECHO = NONE",
    "TITLE = pyNastran for generating solverinput for for Mystran",
    "SUBCASE 1",
    "  SUBTITLE = Default",
    "  LOAD = 1",
    "  SPC = 1",
    "  SPCFORCES(SORT1,REAL) = ALL",
    "  STRESS(SORT1,REAL,VONMISES,BILIN) = ALL",
    "  DISPLACEMENT(SORT1,REAL) = ALL",
])
model.case_control_deck = cc
# model.validate()  # creates an error
"""


##  @}



# ***************************************************************************
# *   Copyright (c) 2014 Juergen Riegel <FreeCAD@juergen-riegel.net>        *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import pathlib
import FreeCAD, Part
from PySide import QtCore
from fcgear import involute
from fcgear import fcgear

if FreeCAD.GuiUp:
    import FreeCADGui
    from PySide import QtGui

__title__ = "PartDesign InvoluteGearObject management"
__author__ = "Juergen Riegel"
__url__ = "https://www.freecad.org"


def makeInvoluteGear(name):
    """makeInvoluteGear(name): makes an InvoluteGear"""
    obj = FreeCAD.ActiveDocument.addObject("Part::Part2DObjectPython", name)
    _InvoluteGear(obj)
    if FreeCAD.GuiUp:
        _ViewProviderInvoluteGear(obj.ViewObject)
    # FreeCAD.ActiveDocument.recompute()
    if FreeCAD.GuiUp:
        body = FreeCADGui.ActiveDocument.ActiveView.getActiveObject("pdbody")
        part = FreeCADGui.ActiveDocument.ActiveView.getActiveObject("part")
        if body:
            body.Group = body.Group + [obj]
        elif part:
            part.Group = part.Group + [obj]
    return obj


class CommandInvoluteGear:
    "GUI command to create an InvoluteGear"

    def GetResources(self):
        return {
            "Pixmap": "PartDesign_InternalExternalGear",
            "MenuText": QtCore.QT_TRANSLATE_NOOP("PartDesign_InvoluteGear", "Involute Gear"),
            "Accel": "",
            "ToolTip": QtCore.QT_TRANSLATE_NOOP(
                "PartDesign_InvoluteGear", "Creates or edits the involute gear definition"
            ),
        }

    def Activated(self):
        FreeCAD.ActiveDocument.openTransaction("Create involute gear")
        FreeCADGui.addModule("InvoluteGearFeature")
        FreeCADGui.doCommand("InvoluteGearFeature.makeInvoluteGear('InvoluteGear')")
        FreeCADGui.doCommand("Gui.activeDocument().setEdit(App.ActiveDocument.ActiveObject.Name,0)")

    def IsActive(self):
        if FreeCAD.ActiveDocument:
            return True
        else:
            return False


class _InvoluteGear:
    "The InvoluteGear object"

    def __init__(self, obj):
        self.Type = "InvoluteGear"
        self._ensure_properties(obj, is_restore=False)
        obj.Proxy = self

    def onDocumentRestored(self, obj):
        """hook used to migrate older versions of this object"""
        self._ensure_properties(obj, is_restore=True)

    def _ensure_properties(self, obj, is_restore):
        def ensure_property(type_, name, doc, default):
            if not hasattr(obj, name):
                obj.addProperty(type_, name, "Gear", doc, locked=True)
                if callable(default):
                    setattr(obj, name, default())
                else:
                    setattr(obj, name, default)

        # for details about the property's docstring translation,
        # see https://tracker.freecad.org/view.php?id=2524
        ensure_property(
            "App::PropertyInteger",
            "NumberOfTeeth",
            doc=QtCore.QT_TRANSLATE_NOOP("App::Property", "Number of gear teeth"),
            default=26,
        )
        ensure_property(
            "App::PropertyLength",
            "Modules",
            doc=QtCore.QT_TRANSLATE_NOOP("App::Property", "Module of the gear"),
            default="2.5 mm",
        )
        ensure_property(
            "App::PropertyAngle",
            "PressureAngle",
            doc=QtCore.QT_TRANSLATE_NOOP("App::Property", "Pressure angle of gear teeth"),
            default="20 deg",
        )
        ensure_property(
            "App::PropertyBool",
            "HighPrecision",
            doc=QtCore.QT_TRANSLATE_NOOP(
                "App::Property",
                "True=2 curves with each 3 control points, False=1 curve with 4 control points.",
            ),
            default=True,
        )
        ensure_property(
            "App::PropertyBool",
            "ExternalGear",
            doc=QtCore.QT_TRANSLATE_NOOP(
                "App::Property", "True=external Gear, False=internal Gear"
            ),
            default=True,
        )
        ensure_property(
            "App::PropertyFloat",
            "AddendumCoefficient",
            doc=QtCore.QT_TRANSLATE_NOOP(
                "App::Property",
                "The height of the tooth from the pitch circle up to its tip, normalized by the module.",
            ),
            default=lambda: 1.0 if obj.ExternalGear else 0.6,
        )
        ensure_property(
            "App::PropertyFloat",
            "DedendumCoefficient",
            doc=QtCore.QT_TRANSLATE_NOOP(
                "App::Property",
                "The height of the tooth from the pitch circle down to its root, normalized by the module.",
            ),
            default=1.25,
        )
        ensure_property(
            "App::PropertyFloat",
            "RootFilletCoefficient",
            doc=QtCore.QT_TRANSLATE_NOOP(
                "App::Property",
                "The radius of the fillet at the root of the tooth, normalized by the module.",
            ),
            default=lambda: 0.375 if is_restore else 0.38,
        )
        ensure_property(
            "App::PropertyFloat",
            "ProfileShiftCoefficient",
            doc=QtCore.QT_TRANSLATE_NOOP(
                "App::Property",
                "The distance by which the reference profile is shifted outwards, normalized by the module.",
            ),
            default=0.0,
        )

    def execute(self, obj):
        w = fcgear.FCWireBuilder()
        generator_func = (
            involute.CreateExternalGear if obj.ExternalGear else involute.CreateInternalGear
        )
        generator_func(
            w,
            obj.Modules.Value,
            obj.NumberOfTeeth,
            obj.PressureAngle.Value,
            split=obj.HighPrecision,
            addCoeff=obj.AddendumCoefficient,
            dedCoeff=obj.DedendumCoefficient,
            filletCoeff=obj.RootFilletCoefficient,
            shiftCoeff=obj.ProfileShiftCoefficient,
        )
        gearw = Part.Wire([o.toShape() for o in w.wire])
        obj.Shape = gearw
        obj.positionBySupport()
        return


class _ViewProviderInvoluteGear:
    "A View Provider for the InvoluteGear object"

    def __init__(self, vobj):
        vobj.Proxy = self

    def getIcon(self):
        return ":/icons/PartDesign_InternalExternalGear.svg"

    def attach(self, vobj):
        self.ViewObject = vobj
        self.Object = vobj.Object

    def setEdit(self, vobj, mode):
        taskd = _InvoluteGearTaskPanel(self.Object, mode)
        taskd.obj = vobj.Object
        taskd.update()
        FreeCADGui.Control.showDialog(taskd)
        return True

    def unsetEdit(self, vobj, mode):
        FreeCADGui.Control.closeDialog()
        return

    def dumps(self):
        return None

    def loads(self, state):
        return None


class _InvoluteGearTaskPanel:
    """The editmode TaskPanel for InvoluteGear objects"""

    def __init__(self, obj, mode):
        self.obj = obj

        self.form = FreeCADGui.PySideUic.loadUi(str(pathlib.Path(__file__).with_suffix(".ui")))
        self.form.setWindowIcon(QtGui.QIcon(":/icons/PartDesign_InternalExternalGear.svg"))
        self.assignToolTipsFromPropertyDocs()

        def assignValue(property_name, fitView=False):
            """Returns a function that takes a single value and assigns it to the given property"""

            def assigner(value):
                setattr(self.obj, property_name, value)
                self.obj.Proxy.execute(self.obj)
                if fitView:
                    FreeCAD.Gui.SendMsgToActiveView("ViewFit")

            return assigner

        def assignIndexAsBool(property_name):
            """Variant of assignValue that transforms the index of a Yes/No Combobox to a bool."""
            assigner = assignValue(property_name)

            def transformingAssigner(value):
                assigner(True if value == 0 else False)

            return transformingAssigner

        self.form.Quantity_Modules.valueChanged.connect(assignValue("Modules", fitView=True))
        self.form.Quantity_PressureAngle.valueChanged.connect(assignValue("PressureAngle"))
        self.form.spinBox_NumberOfTeeth.valueChanged.connect(
            assignValue("NumberOfTeeth", fitView=True)
        )
        self.form.comboBox_HighPrecision.currentIndexChanged.connect(
            assignIndexAsBool("HighPrecision")
        )
        self.form.comboBox_ExternalGear.currentIndexChanged.connect(
            assignIndexAsBool("ExternalGear")
        )
        self.form.doubleSpinBox_Addendum.valueChanged.connect(assignValue("AddendumCoefficient"))
        self.form.doubleSpinBox_Dedendum.valueChanged.connect(assignValue("DedendumCoefficient"))
        self.form.doubleSpinBox_RootFillet.valueChanged.connect(
            assignValue("RootFilletCoefficient")
        )
        self.form.doubleSpinBox_ProfileShift.valueChanged.connect(
            assignValue("ProfileShiftCoefficient")
        )

        self.update()

        if mode == 0:  # fresh created
            self.obj.Proxy.execute(self.obj)  # calculate once
            FreeCAD.Gui.SendMsgToActiveView("ViewFit")

    def assignToolTipsFromPropertyDocs(self):
        def assign(property_name, *widgets):
            doc = self.obj.getDocumentationOfProperty(property_name)
            translated_doc = QtGui.QApplication.translate("App::Property", doc)
            for w in widgets:
                w.setToolTip(translated_doc)

        # we assign the tool tip to both, the label and the input field, for user convenience
        assign("Modules", self.form.Quantity_Modules, self.form.label_Modules)
        assign("PressureAngle", self.form.Quantity_PressureAngle, self.form.label_PressureAngle)
        assign("NumberOfTeeth", self.form.spinBox_NumberOfTeeth, self.form.label_NumberOfTeeth)
        assign("HighPrecision", self.form.comboBox_HighPrecision, self.form.label_HighPrecision)
        assign("ExternalGear", self.form.comboBox_ExternalGear, self.form.label_ExternalGear)
        assign("AddendumCoefficient", self.form.doubleSpinBox_Addendum, self.form.label_Addendum)
        assign("DedendumCoefficient", self.form.doubleSpinBox_Dedendum, self.form.label_Dedendum)
        assign(
            "RootFilletCoefficient", self.form.doubleSpinBox_RootFillet, self.form.label_RootFillet
        )
        assign(
            "ProfileShiftCoefficient",
            self.form.doubleSpinBox_ProfileShift,
            self.form.label_ProfileShift,
        )

    def changeEvent(self, event):
        if event == QtCore.QEvent.LanguageChange:
            self.assignToolTipsFromPropertyDocs()

    def transferTo(self):
        "Transfer from the dialog to the object"
        self.obj.NumberOfTeeth = self.form.spinBox_NumberOfTeeth.value()
        self.obj.Modules = self.form.Quantity_Modules.text()
        self.obj.PressureAngle = self.form.Quantity_PressureAngle.text()
        self.obj.HighPrecision = (
            True if self.form.comboBox_HighPrecision.currentIndex() == 0 else False
        )
        self.obj.ExternalGear = (
            True if self.form.comboBox_ExternalGear.currentIndex() == 0 else False
        )
        self.obj.AddendumCoefficient = self.form.doubleSpinBox_Addendum.value()
        self.obj.DedendumCoefficient = self.form.doubleSpinBox_Dedendum.value()
        self.obj.RootFilletCoefficient = self.form.doubleSpinBox_RootFillet.value()
        self.obj.ProfileShiftCoefficient = self.form.doubleSpinBox_ProfileShift.value()

    def transferFrom(self):
        "Transfer from the object to the dialog"
        self.form.spinBox_NumberOfTeeth.setValue(self.obj.NumberOfTeeth)
        self.form.Quantity_Modules.setText(self.obj.Modules.UserString)
        self.form.Quantity_PressureAngle.setText(self.obj.PressureAngle.UserString)
        self.form.comboBox_HighPrecision.setCurrentIndex(0 if self.obj.HighPrecision else 1)
        self.form.comboBox_ExternalGear.setCurrentIndex(0 if self.obj.ExternalGear else 1)
        self.form.doubleSpinBox_Addendum.setValue(self.obj.AddendumCoefficient)
        self.form.doubleSpinBox_Dedendum.setValue(self.obj.DedendumCoefficient)
        self.form.doubleSpinBox_RootFillet.setValue(self.obj.RootFilletCoefficient)
        self.form.doubleSpinBox_ProfileShift.setValue(self.obj.ProfileShiftCoefficient)

    def getStandardButtons(self):
        return (
            QtGui.QDialogButtonBox.Ok | QtGui.QDialogButtonBox.Cancel | QtGui.QDialogButtonBox.Apply
        )

    def clicked(self, button):
        if button == QtGui.QDialogButtonBox.Apply:
            self.transferTo()
            self.obj.Proxy.execute(self.obj)

    def update(self):
        "fills the widgets"
        self.transferFrom()

    def accept(self):
        self.transferTo()
        FreeCAD.ActiveDocument.recompute()
        FreeCADGui.ActiveDocument.resetEdit()

    def reject(self):
        FreeCADGui.ActiveDocument.resetEdit()
        FreeCAD.ActiveDocument.abortTransaction()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
import abc
from abc import ABC
from typing import Mapping, List, Optional, Tuple, Type
from .uri import AssetUri
from .asset import Asset


class AssetSerializer(ABC):
    for_class: Type[Asset]
    extensions: Tuple[str, ...] = tuple()
    mime_type: str
    can_import: bool = True
    can_export: bool = True

    @classmethod
    @abc.abstractmethod
    def get_label(cls) -> str:
        pass

    @classmethod
    @abc.abstractmethod
    def extract_dependencies(cls, data: bytes) -> List[AssetUri]:
        """Extracts URIs of dependencies from serialized data."""
        pass

    @classmethod
    @abc.abstractmethod
    def serialize(cls, asset: Asset) -> bytes:
        """Serializes an asset object into bytes."""
        pass

    @classmethod
    @abc.abstractmethod
    def deserialize(
        cls,
        data: bytes,
        id: str,
        dependencies: Optional[Mapping[AssetUri, Asset]],
    ) -> "Asset":
        """
        Creates an asset object from serialized data and resolved dependencies.
        If dependencies is None, it indicates a shallow load where dependencies
        were not resolved.
        """
        pass

    @classmethod
    @abc.abstractmethod
    def deep_deserialize(cls, data: bytes) -> Asset:
        """
        Like deserialize(), but builds dependencies itself if they are
        sufficiently defined in the data.

        This method is used for export/import, where some dependencies
        may be embedded in the data, while others may not.
        """
        pass


class DummyAssetSerializer(AssetSerializer):
    """
    A serializer that does nothing. Can be used by simple assets that don't
    need a non-native serialization. These type of assets can implement
    extract_dependencies(), to_bytes() and from_bytes() methods that ignore
    the given serializer.
    """

    @classmethod
    def extract_dependencies(cls, data: bytes) -> List[AssetUri]:
        return []

    @classmethod
    def serialize(cls, asset: Asset) -> bytes:
        return b""

    @classmethod
    def deserialize(
        cls,
        data: bytes,
        id: str,
        dependencies: Optional[Mapping[AssetUri, Asset]],
    ) -> Asset:
        raise RuntimeError("DummySerializer.deserialize() was called")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2017 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""The BIM TogglePanels command"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP

PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")


class BIM_TogglePanels:

    def GetResources(self):
        return {
            "Pixmap": "BIM_TogglePanels",
            "MenuText": QT_TRANSLATE_NOOP("BIM_TogglePanels", "Toggle Bottom Panels"),
            "ToolTip": QT_TRANSLATE_NOOP("BIM_TogglePanels", "Toggles bottom dock panels on/off"),
            "Accel": "Ctrl+0",
        }

    def Activated(self):
        from PySide import QtCore, QtGui

        mw = FreeCADGui.getMainWindow()
        togglebutton = None
        st = mw.statusBar()
        statuswidget = st.findChild(QtGui.QToolBar, "BIMStatusWidget")
        if statuswidget:
            if hasattr(statuswidget, "togglebutton"):
                togglebutton = statuswidget.togglebutton
        dockwidgets = mw.findChildren(QtGui.QDockWidget)
        bottomwidgets = [
            w
            for w in dockwidgets
            if ((mw.dockWidgetArea(w) == QtCore.Qt.BottomDockWidgetArea) and w.isVisible())
        ]
        if bottomwidgets:
            hidden = ""
            for w in bottomwidgets:
                w.hide()
                hidden += w.objectName() + ";;"
                PARAMS.SetString("HiddenWidgets", hidden)
            if togglebutton:
                togglebutton.setChecked(False)
        else:
            widgets = PARAMS.GetString(
                "HiddenWidgets", "Python console;;Report view;;Selection view;;"
            )
            widgets = [mw.findChild(QtGui.QWidget, w) for w in widgets.split(";;") if w]
            for w in widgets:
                w.show()
            if togglebutton:
                togglebutton.setChecked(True)


FreeCADGui.addCommand("BIM_TogglePanels", BIM_TogglePanels())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
from typing import List, Dict, Optional, Iterable, Type
from ..serializer import AssetSerializer


def make_import_filters(serializers: Iterable[Type[AssetSerializer]]) -> List[str]:
    """
    Generates file dialog filters for importing assets.

    Args:
        serializers: A list of AssetSerializer classes.

    Returns:
        A list of filter strings, starting with "All supported files".
    """
    all_extensions = []
    filters = []

    for serializer_class in serializers:
        if not serializer_class.can_import or not serializer_class.extensions:
            continue
        all_extensions.extend(serializer_class.extensions)
        label = serializer_class.get_label()
        extensions = " ".join([f"*{ext}" for ext in serializer_class.extensions])
        filters.append(f"{label} ({extensions})")

    # Add "All supported files" filter if there are any extensions
    if all_extensions:
        combined_extensions = " ".join([f"*{ext}" for ext in sorted(list(set(all_extensions)))])
        filters.insert(0, f"All supported files ({combined_extensions})")

    return filters


def make_export_filters(
    serializers: Iterable[Type[AssetSerializer]],
) -> tuple[List[str], Dict[str, Type[AssetSerializer]]]:
    """
    Generates file dialog filters for exporting assets and a serializer map.

    Args:
        serializers: A list of AssetSerializer classes.

    Returns:
        A tuple of (filters, serializer_map) where filters is a list of filter strings
        starting with "Automatic", and serializer_map maps filter strings to serializers.
    """
    filters = ["Automatic (*)"]
    serializer_map = {}

    for serializer_class in serializers:
        if not serializer_class.can_export or not serializer_class.extensions:
            continue
        label = serializer_class.get_label()
        extensions = " ".join([f"*{ext}" for ext in serializer_class.extensions])
        filter_str = f"{label} ({extensions})"
        filters.append(filter_str)
        serializer_map[filter_str] = serializer_class

    return filters, serializer_map


def get_serializer_from_extension(
    serializers: Iterable[Type[AssetSerializer]],
    file_extension: str,
    for_import: bool | None = None,
) -> Optional[Type[AssetSerializer]]:
    """
    Finds a serializer class based on the file extension and import/export capability.

    Args:
        serializers: A list of AssetSerializer classes.
        file_extension: The file extension (without the leading dot).
        for_import: None = both, True = import, False = export

    Returns:
        The matching AssetSerializer class, or None if not found.
    """
    for_export = for_import is not True
    for_import = for_import is True

    for ser in serializers:
        if for_import and not ser.can_import:
            continue
        if for_export and not ser.can_export:
            continue
        if file_extension in ser.extensions:
            return ser
    return None



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2014 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2014 sliptonic <shopinthewoods@gmail.com>               *
# *   Copyright (c) 2022 - 2025 Larry Woestman <LarryWoestman2@gmail.com>   *
# *   Copyright (c) 2024 Ondsel <development@ondsel.com>                    *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""
The base classes for post processors in the CAM workbench.
"""
import argparse
import importlib.util
import os
from PySide import QtCore, QtGui
import re
import sys
from typing import Any, Dict, List, Optional, Tuple, Union

import Path.Base.Util as PathUtil
import Path.Post.UtilsArguments as PostUtilsArguments
import Path.Post.UtilsExport as PostUtilsExport

import FreeCAD
import Path

translate = FreeCAD.Qt.translate

Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())

debug = False
if debug:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


class _TempObject:
    Path = None
    Name = "Fixture"
    InList = []
    Label = "Fixture"


#
# Define some types that are used throughout this file.
#
Defaults = Dict[str, bool]
FormatHelp = str
GCodeOrNone = Optional[str]
GCodeSections = List[Tuple[str, GCodeOrNone]]
Parser = argparse.ArgumentParser
ParserArgs = Union[None, str, argparse.Namespace]
Postables = Union[List, List[Tuple[str, List]]]
Section = Tuple[str, List]
Sublist = List
Units = str
Values = Dict[str, Any]
Visible = Dict[str, bool]


class PostProcessorFactory:
    """Factory class for creating post processors."""

    @staticmethod
    def get_post_processor(job, postname):
        # Log initial debug message
        Path.Log.debug("PostProcessorFactory.get_post_processor()")

        # Posts have to be in a place we can find them
        paths = Path.Preferences.searchPathsPost()
        paths.extend(sys.path)

        module_name = f"{postname}_post"
        class_name = postname.title()

        # Iterate all the paths to find the module
        for path in paths:
            module_path = os.path.join(path, f"{module_name}.py")
            spec = importlib.util.spec_from_file_location(module_name, module_path)

            if spec and spec.loader:
                module = importlib.util.module_from_spec(spec)
                try:
                    spec.loader.exec_module(module)
                    Path.Log.debug(f"found module {module_name} at {module_path}")

                except (FileNotFoundError, ImportError, ModuleNotFoundError):
                    continue

                try:
                    PostClass = getattr(module, class_name)
                    return PostClass(job)
                except AttributeError:
                    # Return an instance of WrapperPost if no valid module is found
                    Path.Log.debug(f"Post processor {postname} is a script")
                    return WrapperPost(job, module_path, module_name)


def needsTcOp(oldTc, newTc):
    return (
        oldTc is None
        or oldTc.ToolNumber != newTc.ToolNumber
        or oldTc.SpindleSpeed != newTc.SpindleSpeed
        or oldTc.SpindleDir != newTc.SpindleDir
    )


class PostProcessor:
    """Base Class.  All non-legacy postprocessors should inherit from this class."""

    def __init__(self, job, tooltip, tooltipargs, units, *args, **kwargs):
        self._tooltip = tooltip
        self._tooltipargs = tooltipargs
        self._units = units
        self._job = job
        self._args = args
        self._kwargs = kwargs
        self.reinitialize()

    @classmethod
    def exists(cls, processor):
        return processor in Path.Preferences.allAvailablePostProcessors()

    @property
    def tooltip(self):
        """Get the tooltip text for the post processor."""
        raise NotImplementedError("Subclass must implement abstract method")
        # return self._tooltip

    @property
    def tooltipArgs(self) -> FormatHelp:
        return self.parser.format_help()

    @property
    def units(self):
        """Get the units used by the post processor."""
        return self._units

    def _buildPostList(self):
        """
        determines the specific objects and order to postprocess
        Returns a list of objects which can be passed to
        exportObjectsWith() for final posting."""

        def __fixtureSetup(order, fixture, job):
            """Convert a Fixture setting to _TempObject instance with a G0 move to a
            safe height every time the fixture coordinate system change.  Skip
            the move for first fixture, to avoid moving before tool and tool
            height compensation is enabled."""

            fobj = _TempObject()
            c1 = Path.Command(fixture)
            fobj.Path = Path.Path([c1])
            # Avoid any tool move after G49 in preamble and before tool change
            # and G43 in case tool height compensation is in use, to avoid
            # dangerous move without tool compensation.
            if order != 0:
                c2 = Path.Command(
                    "G0 Z"
                    + str(
                        job.Stock.Shape.BoundBox.ZMax + job.SetupSheet.ClearanceHeightOffset.Value
                    )
                )
                fobj.Path.addCommands(c2)
            fobj.InList.append(job)
            return fobj

        wcslist = self._job.Fixtures
        orderby = self._job.OrderOutputBy
        Path.Log.debug(f"Ordering by {orderby}")

        postlist = []

        if orderby == "Fixture":
            Path.Log.debug("Ordering by Fixture")
            # Order by fixture means all operations and tool changes will be
            # completed in one fixture before moving to the next.

            currTc = None
            for index, f in enumerate(wcslist):
                # create an object to serve as the fixture path
                sublist = [__fixtureSetup(index, f, self._job)]

                # Now generate the gcode
                for obj in self._job.Operations.Group:
                    tc = PathUtil.toolControllerForOp(obj)
                    if tc is not None and PathUtil.activeForOp(obj):
                        if needsTcOp(currTc, tc):
                            sublist.append(tc)
                            Path.Log.debug(f"Appending TC: {tc.Name}")
                            currTc = tc
                    sublist.append(obj)
                postlist.append((f, sublist))

        elif orderby == "Tool":
            Path.Log.debug("Ordering by Tool")
            # Order by tool means tool changes are minimized.
            # all operations with the current tool are processed in the current
            # fixture before moving to the next fixture.

            toolstring = "None"
            currTc = None

            # Build the fixture list
            fixturelist = []
            for index, f in enumerate(wcslist):
                # create an object to serve as the fixture path
                fixturelist.append(__fixtureSetup(index, f, self._job))

            # Now generate the gcode
            curlist = []  # list of ops for tool, will repeat for each fixture
            sublist = []  # list of ops for output splitting

            def commitToPostlist():
                if len(curlist) > 0:
                    for fixture in fixturelist:
                        sublist.append(fixture)
                        sublist.extend(curlist)
                    postlist.append((toolstring, sublist))

            Path.Log.track(self._job.PostProcessorOutputFile)
            for idx, obj in enumerate(self._job.Operations.Group):
                Path.Log.track(obj.Label)

                # check if the operation is active
                if not PathUtil.activeForOp(obj):
                    Path.Log.track()
                    continue

                tc = PathUtil.toolControllerForOp(obj)

                # The operation has no ToolController or uses the same
                # ToolController as the previous operations

                if tc is None or not needsTcOp(currTc, tc):
                    # Queue current operation
                    curlist.append(obj)

                # The operation is the first operation or uses a different
                # ToolController as the previous operations

                else:
                    # Commit previous operations
                    commitToPostlist()

                    # Queue current ToolController and operation
                    sublist = [tc]
                    curlist = [obj]
                    currTc = tc

                    # Determine the proper string for the operation's
                    # ToolController
                    if "%T" in self._job.PostProcessorOutputFile:
                        toolstring = f"{tc.ToolNumber}"
                    else:
                        toolstring = re.sub(r"[^\w\d-]", "_", tc.Label)

            # Commit remaining operations
            commitToPostlist()

        elif orderby == "Operation":
            Path.Log.debug("Ordering by Operation")
            # Order by operation means ops are done in each fixture in
            # sequence.
            currTc = None

            # Now generate the gcode
            for obj in self._job.Operations.Group:

                # check if the operation is active
                if not PathUtil.activeForOp(obj):
                    continue

                sublist = []
                Path.Log.debug(f"obj: {obj.Name}")

                for index, f in enumerate(wcslist):
                    sublist.append(__fixtureSetup(index, f, self._job))
                    tc = PathUtil.toolControllerForOp(obj)
                    if tc is not None:
                        if self._job.SplitOutput or needsTcOp(currTc, tc):
                            sublist.append(tc)
                            currTc = tc
                    sublist.append(obj)
                postlist.append((obj.Label, sublist))

        Path.Log.debug(f"Postlist: {postlist}")

        if self._job.SplitOutput:
            Path.Log.track()
            return postlist

        Path.Log.track()
        finalpostlist = [("allitems", [item for slist in postlist for item in slist[1]])]
        Path.Log.debug(f"Postlist: {postlist}")
        return finalpostlist

    def export(self) -> Union[None, GCodeSections]:
        """Process the parser arguments, then postprocess the 'postables'."""
        args: ParserArgs
        flag: bool

        Path.Log.debug("Exporting the job")

        (flag, args) = self.process_arguments()
        #
        # If the flag is True, then continue postprocessing the 'postables'.
        #
        if flag:
            return self.process_postables()
        #
        # The flag is False meaning something unusual happened.
        #
        # If args is None then there was an error during argument processing.
        #
        if args is None:
            return None
        #
        # Otherwise args will contain the argument list formatted for output
        # instead of the "usual" gcode.
        #
        return [("allitems", args)]  # type: ignore

    def init_arguments(
        self,
        values: Values,
        argument_defaults: Defaults,
        arguments_visible: Visible,
    ) -> Parser:
        """Initialize the shared argument definitions."""
        _parser: Parser = PostUtilsArguments.init_shared_arguments(
            values, argument_defaults, arguments_visible
        )
        #
        # Add any argument definitions that are not shared with other postprocessors here.
        #
        return _parser

    def init_argument_defaults(self, argument_defaults: Defaults) -> None:
        """Initialize which arguments (in a pair) are shown as the default argument."""
        PostUtilsArguments.init_argument_defaults(argument_defaults)
        #
        # Modify which argument to show as the default in flag-type arguments here.
        # If the value is True, the first argument will be shown as the default.
        # If the value is False, the second argument will be shown as the default.
        #
        # For example, if you want to show Metric mode as the default, use:
        #   argument_defaults["metric_inch"] = True
        #
        # If you want to show that "Don't pop up editor for writing output" is
        # the default, use:
        #   argument_defaults["show-editor"] = False.
        #
        # Note:  You also need to modify the corresponding entries in the "values" hash
        #        to actually make the default value(s) change to match.
        #

    def init_arguments_visible(self, arguments_visible: Visible) -> None:
        """Initialize which argument pairs are visible in TOOLTIP_ARGS."""
        PostUtilsArguments.init_arguments_visible(arguments_visible)
        #
        # Modify the visibility of any arguments from the defaults here.
        #

    def init_values(self, values: Values) -> None:
        """Initialize values that are used throughout the postprocessor."""
        #
        PostUtilsArguments.init_shared_values(values)
        #
        # Set any values here that need to override the default values set
        # in the init_shared_values routine.
        #
        values["UNITS"] = self._units

    def process_arguments(self) -> Tuple[bool, ParserArgs]:
        """Process any arguments to the postprocessor."""
        #
        # This function is separated out to make it easier to inherit from this class.
        #
        args: ParserArgs
        flag: bool

        (flag, args) = PostUtilsArguments.process_shared_arguments(
            self.values, self.parser, self._job.PostProcessorArgs, self.all_visible, "-"
        )
        #
        # If the flag is True, then all of the arguments should be processed normally.
        #
        if flag:
            #
            # Process any additional arguments here.
            #
            #
            # Update any variables that might have been modified while processing the arguments.
            #
            self._units = self.values["UNITS"]
        #
        # If the flag is False, then args is either None (indicating an error while
        # processing the arguments) or a string containing the argument list formatted
        # for output.  Either way the calling routine will need to handle the args value.
        #
        return (flag, args)

    def process_postables(self) -> GCodeSections:
        """Postprocess the 'postables' in the job to g code sections."""
        #
        # This function is separated out to make it easier to inherit from this class.
        #
        gcode: GCodeOrNone
        g_code_sections: GCodeSections
        partname: str
        postables: Postables
        section: Section
        sublist: Sublist

        postables = self._buildPostList()

        Path.Log.debug(f"postables count: {len(postables)}")

        g_code_sections = []
        for _, section in enumerate(postables):
            partname, sublist = section
            gcode = PostUtilsExport.export_common(self.values, sublist, "-")
            g_code_sections.append((partname, gcode))

        return g_code_sections

    def reinitialize(self) -> None:
        """Initialize or reinitialize the 'core' data structures for the postprocessor."""
        #
        # This is also used to reinitialize the data structures between tests.
        #
        self.values: Values = {}
        self.init_values(self.values)
        self.argument_defaults: Defaults = {}
        self.init_argument_defaults(self.argument_defaults)
        self.arguments_visible: Visible = {}
        self.init_arguments_visible(self.arguments_visible)
        self.parser: Parser = self.init_arguments(
            self.values, self.argument_defaults, self.arguments_visible
        )
        #
        # Create another parser just to get a list of all possible arguments
        # that may be output using --output_all_arguments.
        #
        self.all_arguments_visible: Visible = {}
        for k in iter(self.arguments_visible):
            self.all_arguments_visible[k] = True
        self.all_visible: Parser = self.init_arguments(
            self.values, self.argument_defaults, self.all_arguments_visible
        )


class WrapperPost(PostProcessor):
    """Wrapper class for old post processors that are scripts."""

    def __init__(self, job, script_path, module_name, *args, **kwargs):
        super().__init__(job, tooltip=None, tooltipargs=None, units=None, *args, **kwargs)
        self.script_path = script_path
        self.module_name = module_name
        Path.Log.debug(f"WrapperPost.__init__({script_path})")
        self.load_script()

    def load_script(self):
        """Dynamically load the script as a module."""
        try:
            spec = importlib.util.spec_from_file_location(self.module_name, self.script_path)
            self.script_module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(self.script_module)
        except Exception as e:
            raise ImportError(f"Failed to load script: {e}")

        if not hasattr(self.script_module, "export"):
            raise AttributeError("The script does not have an 'export' function.")

        # Set properties based on attributes of the module
        self._units = "Metric" if getattr(self.script_module, "UNITS", "G21") == "G21" else "Inch"
        self._tooltip = getattr(self.script_module, "TOOLTIP", "No tooltip provided")
        self._tooltipargs = getattr(self.script_module, "TOOLTIP_ARGS", [])

    def export(self):
        """Dynamically reload the module for the export to ensure up-to-date usage."""

        postables = self._buildPostList()
        Path.Log.debug(f"postables count: {len(postables)}")

        g_code_sections = []
        for idx, section in enumerate(postables):
            partname, sublist = section

            gcode = self.script_module.export(sublist, "-", self._job.PostProcessorArgs)
            Path.Log.debug(f"Exported {partname}")
            g_code_sections.append((partname, gcode))
        return g_code_sections

    @property
    def tooltip(self):
        return self._tooltip

    @property
    def tooltipArgs(self):
        return self._tooltipargs

    @property
    def units(self):
        return self._units
